import base64
import json
import websocket


class WebsocketClient(object):
    def __init__(self, url, username=None, password=None, headers={}, enable_trace=False,
                 timeout=60):
        """Simple Web socket client for wrapanapi

       Args:
           url: String with the hostname or IP address of the server with port
            (e.g. 'ws://10.11.12.13:8080/hawkular/command-gateway/ui/ws')
           username: username for basic auth (optional)
           password: password for basic auth
           headers: When you want to pass specified header use this
           enable_trace: Enable trace on web socket client
           timeout: receive timeout in seconds
       """
        self.url = url
        self.username = username
        self.password = password
        self.headers = headers
        self.enable_trace = enable_trace
        self.timeout = timeout
        self.ws = None

    def connect(self):
        """connects with the initialized detail"""
        if self.ws and self.ws.connected:
            return self.ws
        else:
            websocket.enableTrace(self.enable_trace)
            if self.username:
                base64_creds = base64.b64encode("{}:{}".format(self.username, self.password))
                self.headers.update({"Authorization": "Basic {}".format(base64_creds)})
            self.ws = websocket.create_connection(self.url, header=self.headers)
            self.ws.settimeout(self.timeout)

    @property
    def connected(self):
        """Returns True if client connected with server"""
        if self.ws and self.ws.connected:
            return True
        return False

    def close(self):
        """Close the current connection"""
        if self.ws and self.ws.connected:
            self.ws.close()

    def _check_connection(self, make_connection=True):
        """Check client connected with server

        Args:
            make_connection: by default True. Creates connection if not connected
        """
        if self.connected:
            return
        elif make_connection:
            self.connect()
            if self.connected:
                return
        raise RuntimeError("WS client not connected!")

    def send(self, payload, binary_stream=False):
        """Send payload to server
        Args:
            payload: payload, string or binary
            binary_stream: dafault False. Set this True when you send binary payload
        """
        self._check_connection()
        if binary_stream:
            self.ws.send_binary(payload=payload)
        else:
            self.ws.send(payload=payload)

    def receive(self):
        """Returns available message on received queue. If there is no message.
         waits till timeout"""
        self._check_connection()
        return self.ws.recv()


class HawkularWebsocketClient(WebsocketClient):
    """This client extended from normal websocket client. designed to hawkular specific"""
    def __init__(self, url, username=None, password=None, headers={}, enable_trace=False,
                 timeout=60):
        """Creates hawkular web socket client. for arguments refer 'WebsocketClient'"""
        super(HawkularWebsocketClient, self).__init__(url=url, username=username, password=password,
                                                      headers=headers, enable_trace=enable_trace,
                                                      timeout=timeout)
        self.session_id = None

    def connect(self):
        """Create connection with hawkular web socket server"""
        super(HawkularWebsocketClient, self).connect()
        response = self.hwk_receive()
        if 'WelcomeResponse' in response:
            self.session_id = response['WelcomeResponse']['sessionId']
            return response['WelcomeResponse']
        else:
            raise RuntimeWarning("Key 'WelcomeResponse' not found on response: {}".format(response))
            return response

    def hwk_receive(self):
        """parse recevied message and returns as dictionary value"""
        payload = self.receive()
        data = payload.split('=', 1)
        if len(data) != 2:
            raise IndentationError("Unknown payload format! {}".format(payload))
        response = {data[0]: json.loads(data[1])}
        if 'GenericErrorResponse' in response:
            raise Exception("Hawkular server sent failure message: {}"
                            .format(response['GenericErrorResponse']))
        return response

    def hwk_invoke_operation(self, payload, operation_name="ExecuteOperation", binary_content=None,
                             binary_file_location=None, wait_for_response=True):
        """Runs hawkular specific operations
        Args:
            payload: payload to server. only string
            operation_name: requested operation. default: ExecuteOperation
            binary_content: binary content
            binary_file_location: binary content file name. Will be changed as binary content
            wait_for_response: When executing a command, wait for the response. default: True

        """
        _payload = "{}Request={}".format(operation_name, json.dumps(payload))
        if binary_file_location:
            binary_content = open(binary_file_location, 'rb').read()
        if binary_content:
            self.send(_payload + binary_content, binary_stream=True)
        else:
            self.send(_payload, binary_stream=False)
        if wait_for_response:
            responses = []
            response = self.hwk_receive()
            responses.append(response)
            if "GenericSuccessResponse" in response:
                responses.append(self.hwk_receive())
            return responses



# coding: utf-8
"""Backend management system classes

Used to communicate with providers without using CFME facilities
"""
from contextlib import contextmanager
from datetime import datetime
from functools import partial

from cinderclient.v2 import client as cinderclient
from cinderclient import exceptions as cinder_exceptions
from heatclient import client as heat_client
from keystoneclient.v2_0 import client as oskclient
from novaclient import client as osclient
from novaclient import exceptions as os_exceptions
from novaclient.client import HTTPClient
from novaclient.v2.floating_ips import FloatingIP
from novaclient.v2.servers import Server
from requests.exceptions import Timeout
import json
import time
import tzlocal
from wait_for import wait_for

from base import MgmtSystemAPIBase, VMInfo
from exceptions import (
    NoMoreFloatingIPs, NetworkNameNotFound, VMInstanceNotFound, VMNotFoundViaIP,
    ActionTimedOutError, VMError
)


# TODO The following monkeypatch nonsense is criminal, and would be
# greatly simplified if openstack made it easier to specify a custom
# client class. This is a trivial PR that they're likely to accept.

# Note: This same mechanism may be required for keystone and cinder
# clients, but hopefully won't be.

# monkeypatch method to add retry support to openstack
def _request_timeout_handler(self, url, method, retry_count=0, **kwargs):
    try:
        # Use the original request method to do the actual work
        return HTTPClient.request(self, url, method, **kwargs)
    except Timeout:
        if retry_count >= 3:
            self._cfme_logger.error('nova request timed out after {} retries'.format(retry_count))
            raise
        else:
            # feed back into the replaced method that supports retry_count
            retry_count += 1
            self._cfme_logger.info('nova request timed out; retry {}'.format(retry_count))
            return self.request(url, method, retry_count=retry_count, **kwargs)


class OpenstackSystem(MgmtSystemAPIBase):
    """Openstack management system

    Uses novaclient.

    Args:
        tenant: The tenant to log in with.
        username: The username to connect with.
        password: The password to connect with.
        auth_url: The authentication url.

    """

    _stats_available = {
        'num_vm': lambda self: len(self._get_all_instances(True)),
        'num_template': lambda self: len(self.list_template()),
    }

    states = {
        'paused': ('PAUSED',),
        'running': ('ACTIVE',),
        'stopped': ('SHUTOFF',),
        'suspended': ('SUSPENDED',),
    }

    can_suspend = True
    can_pause = True

    def __init__(self, **kwargs):
        super(OpenstackSystem, self).__init__(kwargs)
        self.tenant = kwargs['tenant']
        self.username = kwargs['username']
        self.password = kwargs['password']
        self.auth_url = kwargs['auth_url']
        self._api = None
        self._kapi = None
        self._capi = None

    @property
    def api(self):
        if not self._api:
            self._api = osclient.Client('2',
                                        self.username,
                                        self.password,
                                        self.tenant,
                                        self.auth_url,
                                        service_type="compute",
                                        insecure=True,
                                        timeout=30)
            # replace the client request method with our version that
            # can handle timeouts; uses explicit binding (versus
            # replacing the method directly on the HTTPClient class)
            # so we can still call out to HTTPClient's original request
            # method in the timeout handler method
            self._api.client._cfme_logger = self.logger
            self._api.client.request = _request_timeout_handler.__get__(self._api.client,
                HTTPClient)
        return self._api

    @property
    def kapi(self):
        if not self._kapi:
            self._kapi = oskclient.Client(username=self.username,
                                          password=self.password,
                                          tenant_name=self.tenant,
                                          auth_url=self.auth_url,
                                          insecure=True)
        return self._kapi

    @property
    def capi(self):
        if not self._capi:
            self._capi = cinderclient.Client(self.username,
                                             self.password,
                                             self.tenant,
                                             self.auth_url,
                                             service_type="volume",
                                             insecure=True)
        return self._capi

    @property
    def stackapi(self):
        ks_client = oskclient.Client(username=self.username,
                                     password=self.password,
                                     tenant_name=self.tenant,
                                     auth_url=self.auth_url,
                                     insecure=True)
        heat_endpoint = ks_client.service_catalog.url_for(service_type='orchestration',
                                                          endpoint_type='publicURL')
        self._stackapi = heat_client.Client('1', heat_endpoint, token=ks_client.auth_token)
        return self._stackapi

    def _get_tenants(self):
        real_tenants = []
        tenants = self.kapi.tenants.list()
        for tenant in tenants:
            users = tenant.list_users()
            user_list = [user.name for user in users]
            if self.username in user_list:
                real_tenants.append(tenant)
        return real_tenants

    def _get_tenant(self, **kwargs):
        return self.kapi.tenants.find(**kwargs).id

    def _get_user(self, **kwargs):
        return self.kapi.users.find(**kwargs).id

    def _get_role(self, **kwargs):
        return self.kapi.roles.find(**kwargs).id

    def add_tenant(self, tenant_name, description=None, enabled=True, user=None, roles=None):
        tenant = self.kapi.tenants.create(tenant_name=tenant_name,
                                          description=description,
                                          enabled=enabled)
        if user and roles:
            user = self._get_user(name=user)
            for role in roles:
                role_id = self._get_role(name=role)
                tenant.add_user(user, role_id)
        return tenant.id

    def list_tenant(self):
        return [i.name for i in self._get_tenants()]

    def remove_tenant(self, tenant_name):
        tid = self._get_tenant(name=tenant_name)
        self.kapi.tenants.delete(tid)

    def start_vm(self, instance_name):
        self.logger.info(" Starting OpenStack instance %s" % instance_name)
        if self.is_vm_running(instance_name):
            return True

        instance = self._find_instance_by_name(instance_name)
        if self.is_vm_suspended(instance_name):
            instance.resume()
        elif self.is_vm_paused(instance_name):
            instance.unpause()
        else:
            instance.start()
        wait_for(lambda: self.is_vm_running(instance_name), message="start %s" % instance_name)
        return True

    def stop_vm(self, instance_name):
        self.logger.info(" Stopping OpenStack instance %s" % instance_name)
        if self.is_vm_stopped(instance_name):
            return True

        instance = self._find_instance_by_name(instance_name)
        instance.stop()
        wait_for(lambda: self.is_vm_stopped(instance_name), message="stop %s" % instance_name)
        return True

    def create_vm(self):
        raise NotImplementedError('create_vm not implemented.')

    def delete_vm(self, instance_name, delete_fip=True):
        self.logger.info(" Deleting OpenStack instance {}".format(instance_name))
        instance = self._find_instance_by_name(instance_name)
        if delete_fip:
            self.unassign_and_delete_floating_ip(instance)
        else:
            self.unassign_floating_ip(instance)
        self.logger.info(" Deleting OpenStack instance {} in progress now.".format(instance_name))
        instance.delete()
        wait_for(lambda: not self.does_vm_exist(instance_name), timeout='3m', delay=5)
        return True

    def restart_vm(self, instance_name):
        self.logger.info(" Restarting OpenStack instance %s" % instance_name)
        return self.stop_vm(instance_name) and self.start_vm(instance_name)

    def list_vm(self, **kwargs):
        instance_list = self._get_all_instances()
        return [instance.name for instance in instance_list]

    def list_template(self):
        template_list = self.api.images.list()
        return [template.name for template in template_list]

    def list_flavor(self):
        flavor_list = self.api.flavors.list()
        return [flavor.name for flavor in flavor_list]

    def list_volume(self):  # TODO: maybe names? Could not get it to work via API though ...
        volume_list = self.capi.volumes.list()
        return [volume.id for volume in volume_list]

    def list_network(self):
        network_list = self.api.networks.list()
        return [network.label for network in network_list]

    def info(self):
        return '%s %s' % (self.api.client.service_type, self.api.client.version)

    def disconnect(self):
        pass

    def vm_status(self, vm_name):
        """Retrieve Instance status.

        Raises:
            :py:class:`mgmtsystem.exceptions.VMError
        """
        inst = self._find_instance_by_name(vm_name)
        if inst.status != "ERROR":
            return inst.status
        if not hasattr(inst, "fault"):
            raise VMError("Instance {} in error state!".format(vm_name))
        raise VMError("Instance {} error {}: {} | {}".format(
            vm_name, inst.fault["code"], inst.fault["message"], inst.fault["created"]))

    def create_volume(self, size_gb, **kwargs):
        volume = self.capi.volumes.create(size_gb, **kwargs).id
        wait_for(lambda: self.capi.volumes.get(volume).status == "available", num_sec=60, delay=0.5)
        return volume

    def delete_volume(self, *ids, **kwargs):
        wait = kwargs.get("wait", True)
        timeout = kwargs.get("timeout", 180)
        for id in ids:
            self.capi.volumes.find(id=id).delete()
        if not wait:
            return
        # Wait for them
        wait_for(
            lambda: all(map(lambda id: not self.volume_exists(id), ids)),
            delay=0.5, num_sec=timeout)

    def volume_exists(self, id):
        try:
            self.capi.volumes.get(id)
            return True
        except cinder_exceptions.NotFound:
            return False

    def get_volume(self, id):
        return self.capi.volumes.get(id)

    @contextmanager
    def with_volume(self, *args, **kwargs):
        """Creates a context manager that creates a single volume with parameters defined via params
        and destroys it after exiting the context manager

        For arguments description, see the :py:meth:`OpenstackSystem.create_volume`.
        """
        volume = self.create_volume(*args, **kwargs)
        try:
            yield volume
        finally:
            self.delete_volume(volume)

    @contextmanager
    def with_volumes(self, *configurations, **kwargs):
        """Similar to :py:meth:`OpenstackSystem.with_volume`, but with multiple volumes.

        Args:
            *configurations: Can be either :py:class:`int` (taken as a disk size), or a tuple.
                If it is a tuple, then first element is disk size and second element a dictionary
                of kwargs passed to :py:meth:`OpenstackSystem.create_volume`. Can be 1-n tuple, it
                can cope with that.
        Keywords:
            n: How many copies of single configuration produce? Useful when you want to create eg.
                10 identical volumes, so you specify only one configuration and set n=10.

        Example:

            .. code-block:: python

               with mgmt.with_volumes(1, n=10) as (d0, d1, d2, d3, d4, d5, d6, d7, d8, d9):
                   pass  # provisions 10 identical 1G volumes

               with mgmt.with_volumes(1, 2) as (d0, d1):
                   pass  # d0 1G, d1 2G

               with mgmt.with_volumes((1, {}), (2, {})) as (d0, d1):
                   pass  # d0 1G, d1 2G same as before but you can see you can pass kwargs through

        """
        n = kwargs.pop("n", None)
        if n is None:
            pass  # Nothing to do
        elif n > 1 and len(configurations) == 1:
            configurations = n * configurations
        elif n != len(configurations):
            raise "n does not equal the length of configurations"
        # now n == len(configurations)
        volumes = []
        try:
            for configuration in configurations:
                if isinstance(configuration, int):
                    size, kwargs = configuration, {}
                elif len(configuration) == 1:
                    size, kwargs = configuration[0], {}
                elif len(configuration) == 2:
                    size, kwargs = configuration
                else:
                    size = configuration[0]
                    kwargs = configuration[1]
                volumes.append(self.create_volume(size, **kwargs))
            yield volumes
        finally:
            self.delete_volume(*volumes)

    def _get_instance_name(self, id):
        return self.api.servers.get(id).name

    def volume_attachments(self, volume_id):
        """Returns a dictionary of ``{instance: device}`` relationship of the volume."""
        volume = self.capi.volumes.get(volume_id)
        result = {}
        for attachment in volume.attachments:
            result[self._get_instance_name(attachment["server_id"])] = attachment["device"]
        return result

    def vm_creation_time(self, vm_name):
        instance = self._find_instance_by_name(vm_name)
        # Example vm.created: 2014-08-14T23:29:30Z
        create_time = datetime.strptime(instance.created, '%Y-%m-%dT%H:%M:%SZ')
        # create time is UTC, localize it, strip tzinfo
        return tzlocal.get_localzone().fromutc(create_time).replace(tzinfo=None)

    def is_vm_running(self, vm_name):
        return self.vm_status(vm_name) in self.states['running']

    def is_vm_stopped(self, vm_name):
        return self.vm_status(vm_name) in self.states['stopped']

    def is_vm_suspended(self, vm_name):
        return self.vm_status(vm_name) in self.states['suspended']

    def is_vm_paused(self, vm_name):
        return self.vm_status(vm_name) in self.states['paused']

    def wait_vm_running(self, vm_name, num_sec=360):
        self.logger.info(" Waiting for OS instance %s to change status to ACTIVE" % vm_name)
        wait_for(self.is_vm_running, [vm_name], num_sec=num_sec)

    def wait_vm_stopped(self, vm_name, num_sec=360):
        self.logger.info(" Waiting for OS instance %s to change status to SHUTOFF" % vm_name)
        wait_for(self.is_vm_stopped, [vm_name], num_sec=num_sec)

    def wait_vm_suspended(self, vm_name, num_sec=720):
        self.logger.info(" Waiting for OS instance %s to change status to SUSPENDED" % vm_name)
        wait_for(self.is_vm_suspended, [vm_name], num_sec=num_sec)

    def wait_vm_paused(self, vm_name, num_sec=720):
        self.logger.info(" Waiting for OS instance %s to change status to PAUSED" % vm_name)
        wait_for(self.is_vm_paused, [vm_name], num_sec=num_sec)

    def suspend_vm(self, instance_name):
        self.logger.info(" Suspending OpenStack instance %s" % instance_name)
        if self.is_vm_suspended(instance_name):
            return True

        instance = self._find_instance_by_name(instance_name)
        instance.suspend()
        wait_for(lambda: self.is_vm_suspended(instance_name), message="suspend %s" % instance_name)

    def pause_vm(self, instance_name):
        self.logger.info(" Pausing OpenStack instance %s" % instance_name)
        if self.is_vm_paused(instance_name):
            return True

        instance = self._find_instance_by_name(instance_name)
        instance.pause()
        wait_for(lambda: self.is_vm_paused(instance_name), message="pause %s" % instance_name)

    def clone_vm(self, source_name, vm_name):
        raise NotImplementedError('clone_vm not implemented.')

    def free_fips(self, pool):
        """Returns list of free floating IPs sorted by ip address."""
        return sorted(self.api.floating_ips.findall(fixed_ip=None, pool=pool), key=lambda ip: ip.ip)

    def deploy_template(self, template, *args, **kwargs):
        """ Deploys an OpenStack instance from a template.

        For all available args, see ``create`` method found here:
        http://docs.openstack.org/developer/python-novaclient/ref/v1_1/servers.html

        Most important args are listed below.

        Args:
            template: The name of the template to use.
            flavour_name: The name of the flavour to use.
            flavour_id: UUID of the flavour to use.
            vm_name: A name to use for the vm.
            network_name: The name of the network if it is a multi network setup (Havanna).

        Note: If assign_floating_ip kwarg is present, then :py:meth:`OpenstackSystem.create_vm` will
            attempt to register a floating IP address from the pool specified in the arg.
        """
        power_on = kwargs.pop("power_on", True)
        nics = []
        timeout = kwargs.pop('timeout', 900)

        if 'flavour_name' in kwargs:
            flavour = self.api.flavors.find(name=kwargs['flavour_name'])
        elif 'instance_type' in kwargs:
            flavour = self.api.flavors.find(name=kwargs['instance_type'])
        elif 'flavour_id' in kwargs:
            flavour = self.api.flavors.find(id=kwargs['flavour_id'])
        else:
            flavour = self.api.flavors.find(name='m1.tiny')
        if 'vm_name' not in kwargs:
            vm_name = 'new_instance_name'
        else:
            vm_name = kwargs['vm_name']
        self.logger.info(" Deploying OpenStack template %s to instance %s (%s)" % (
            template, kwargs["vm_name"], flavour.name))
        if len(self.list_network()) > 1:
            if 'network_name' not in kwargs:
                raise NetworkNameNotFound('Must select a network name')
            else:
                net_id = self.api.networks.find(label=kwargs['network_name']).id
                nics = [{'net-id': net_id}]

        image = self.api.images.find(name=template)
        instance = self.api.servers.create(vm_name, image, flavour, nics=nics, *args, **kwargs)
        self.wait_vm_running(vm_name, num_sec=timeout)
        if kwargs.get('floating_ip_pool', None):
            self.assign_floating_ip(instance, kwargs['floating_ip_pool'])

        if power_on:
            self.start_vm(vm_name)

        return vm_name

    def assign_floating_ip(self, instance_or_name, floating_ip_pool, safety_timer=5):
        """Assigns a floating IP to an instance.

        Args:
            instance_or_name: Name of the instance or instance object itself.
            floating_ip_pool: Name of the floating IP pool to take from.
            safety_timer: A timeout after assigning the FIP that is used to detect whether another
                external influence did not steal our FIP. Default is 5.

        Returns:
            The public FIP. Raises an exception in case of error.
        """
        instance = self._instance_or_name(instance_or_name)

        current_ip = self.current_ip_address(instance.name)
        if current_ip is not None:
            return current_ip

        # Why while? Well, this code can cause one peculiarity. Race condition can "steal" a FIP
        # so this will loop until it really get the address. A small timeout is added to ensure
        # the instance really got that address and other process did not steal it.
        # TODO: Introduce neutron client and its create+assign?
        while self.current_ip_address(instance.name) is None:
            free_ips = self.free_fips(floating_ip_pool)
            # We maintain 1 floating IP as a protection against race condition
            # I know it is bad practice, but I did not figure out how to prevent the race
            # condition by openstack saying "Hey, this IP is already assigned somewhere"
            if len(free_ips) > 1:
                # There are 2 and more ips, so we will take the first one (eldest)
                ip = free_ips[0]
                self.logger.info("Reusing {} from pool {}".format(ip.ip, floating_ip_pool))
            else:
                # There is one or none, so create one.
                try:
                    ip = self.api.floating_ips.create(floating_ip_pool)
                except (os_exceptions.ClientException, os_exceptions.OverLimit) as e:
                    self.logger.error("Probably no more FIP slots available: {}".format(str(e)))
                    free_ips = self.free_fips(floating_ip_pool)
                    # So, try picking one from the list (there still might be one)
                    if free_ips:
                        # There is something free. Slight risk of race condition
                        ip = free_ips[0]
                        self.logger.info(
                            "Reused {} from pool {} because no more free spaces for new ips"
                            .format(ip.ip, floating_ip_pool))
                    else:
                        # Nothing can be done
                        raise NoMoreFloatingIPs("Provider {} ran out of FIPs".format(self.auth_url))
                self.logger.info("Created {} in pool {}".format(ip.ip, floating_ip_pool))
            instance.add_floating_ip(ip)

            # Now the grace period in which a FIP theft could happen
            time.sleep(safety_timer)

        self.logger.info("Instance {} got a floating IP {}".format(instance.name, ip.ip))
        return self.current_ip_address(instance.name)

    def unassign_floating_ip(self, instance_or_name):
        """Disassociates the floating IP (if present) from VM.

        Args:
            instance_or_name: Name of the instance or instance object itself.

        Returns:
            None if no FIP was dissociated. Otherwise it will return the Floating IP object.
        """
        instance = self._instance_or_name(instance_or_name)
        ip_addr = self.current_ip_address(instance.name)
        if ip_addr is None:
            return None
        floating_ips = self.api.floating_ips.findall(ip=ip_addr)
        if not floating_ips:
            return None
        floating_ip = floating_ips[0]
        self.logger.info(
            'Detaching floating IP {}/{} from {}'.format(
                floating_ip.id, floating_ip.ip, instance.name))
        instance.remove_floating_ip(floating_ip)
        wait_for(
            lambda: self.current_ip_address(instance.name) is None, delay=1, timeout='1m')
        return floating_ip

    def delete_floating_ip(self, floating_ip):
        """Deletes an existing FIP.

        Args:
            floating_ip: FloatingIP object or an IP address of the FIP.

        Returns:
            True if it deleted a FIP, False if it did not delete it, most probably because it
            does not exist.
        """
        if floating_ip is None:
            # To be able to chain with unassign_floating_ip, which can return None
            return False
        if not isinstance(floating_ip, FloatingIP):
            floating_ip = self.api.floating_ips.findall(ip=floating_ip)
            if not floating_ip:
                return False
            floating_ip = floating_ip[0]
        self.logger.info('Deleting floating IP {}/{}'.format(floating_ip.id, floating_ip.ip))
        floating_ip.delete()
        wait_for(
            lambda: len(self.api.floating_ips.findall(ip=floating_ip.ip)) == 0,
            delay=1, timeout='1m')
        return True

    def unassign_and_delete_floating_ip(self, instance_or_name):
        """Disassociates the floating IP (if present) from VM and deletes it.

        Args:
            instance_or_name: Name of the instance or instance object itself.

        Returns:
            True if it deleted a FIP, False if it did not delete it, most probably because it
            does not exist.
        """
        return self.delete_floating_ip(self.unassign_floating_ip(instance_or_name))

    def _get_instance_networks(self, name):
        instance = self._find_instance_by_name(name)
        return instance._info['addresses']

    def current_ip_address(self, name):
        networks = self._get_instance_networks(name)
        for network_nics in networks.itervalues():
            for nic in network_nics:
                if nic['OS-EXT-IPS:type'] == 'floating':
                    return str(nic['addr'])

    def all_vms(self):
        result = []
        for vm in self._get_all_instances():
            ip = None
            for network_nics in vm._info["addresses"].itervalues():
                for nic in network_nics:
                    if nic['OS-EXT-IPS:type'] == 'floating':
                        ip = str(nic['addr'])
            result.append(VMInfo(
                vm.id,
                vm.name,
                vm.status,
                ip,
            ))
        return result

    def get_vm_name_from_ip(self, ip):
        # unfortunately it appears you cannot query for ip address from the sdk,
        #   unlike curling rest api which does work
        """ Gets the name of a vm from its IP.

        Args:
            ip: The ip address of the vm.
        Returns: The vm name for the corresponding IP."""

        instances = self._get_all_instances()

        for instance in instances:
            addr = self.get_ip_address(instance.name)
            if addr is not None and ip in addr:
                return str(instance.name)
        raise VMNotFoundViaIP('The requested IP is not known as a VM')

    def get_ip_address(self, name, **kwargs):
        return self.current_ip_address(name)

    def _generic_paginator(self, f):
        """A generic paginator for OpenStack services

        Takes a callable and recursively runs the "listing" until no more are returned
        by sending the ```marker``` kwarg to offset the search results. We try to rollback
        up to 10 times in the markers in case one was deleted. If we can't rollback after
        10 times, we give up.
        Possible improvement is to roll back in 5s or 10s, but then we have to check for
        uniqueness and do dup removals.
        """
        lists = []
        marker = None
        while True:
            if not lists:
                temp_list = f()
            else:
                for i in range(min(10, len(lists))):
                    list_offset = -(i + 1)
                    marker = lists[list_offset].id
                    try:
                        temp_list = f(marker=marker)
                        break
                    except os_exceptions.BadRequest:
                        continue
                else:
                    raise Exception("Could not get list, maybe mass deletion after 10 marker tries")
            if temp_list:
                lists.extend(temp_list)
            else:
                break
        return lists

    def _get_all_instances(self, filter_tenants=True):
        call = partial(self.api.servers.list, True, {'all_tenants': True})
        instances = self._generic_paginator(call)
        if filter_tenants:
            # Filter instances based on their tenant ID
            # needed for CFME 5.3 and higher
            tenants = self._get_tenants()
            ids = [tenant.id for tenant in tenants]
            instances = filter(lambda i: i.tenant_id in ids, instances)
        return instances

    def _get_all_templates(self):
        return self.api.images.list()

    def _find_instance_by_name(self, name):
        """
        OpenStack Nova Client does have a find method, but it doesn't
        allow the find method to be used on other tenants. The list()
        method is the only one that allows an all_tenants=True keyword
        """
        instances = self._get_all_instances()
        for instance in instances:
            if instance.name == name:
                return instance
        else:
            raise VMInstanceNotFound(name)

    def _instance_or_name(self, instance_or_name):
        """Works similarly to _find_instance_by_name but allows passing the constructed object."""
        if isinstance(instance_or_name, Server):
            # Object passed
            return instance_or_name
        else:
            # String passed
            return self._find_instance_by_name(instance_or_name)

    def _find_template_by_name(self, name):
        templates = self._get_all_templates()
        for template in templates:
            if template.name == name:
                return template
        else:
            raise VMInstanceNotFound("template {}".format(name))

    def get_template_id(self, name):
        return self._find_template_by_name(name).id

    def does_vm_exist(self, name):
        try:
            self._find_instance_by_name(name)
            return True
        except Exception:
            return False

    def remove_host_from_cluster(self, hostname):
        raise NotImplementedError('remove_host_from_cluster not implemented')

    def get_first_floating_ip(self):
        try:
            self.api.floating_ips.create()
        except os_exceptions.NotFound:
            self.logger.error('No more Floating IPs available, will attempt to grab a free one')
        try:
            first_available_ip = (ip for ip in self.api.floating_ips.list()
                                  if ip.instance_id is None).next()
        except StopIteration:
            return None
        return first_available_ip.ip

    def mark_as_template(self, instance_name, **kwargs):
        """OpenStack marking as template is a little bit more complex than vSphere.

        We have to rename the instance, create a snapshot of the original name and then delete the
        instance."""
        self.logger.info("Marking {} as OpenStack template".format(instance_name))
        instance = self._find_instance_by_name(instance_name)
        original_name = instance.name
        copy_name = original_name + "_copytemplate"
        instance.update(copy_name)
        try:
            self.wait_vm_steady(copy_name)
            if not self.is_vm_stopped(copy_name):
                instance.stop()
                self.wait_vm_stopped(copy_name)
            uuid = instance.create_image(original_name)
            wait_for(lambda: self.api.images.get(uuid).status == "ACTIVE", num_sec=900, delay=5)
            instance.delete()
            wait_for(lambda: not self.does_vm_exist(copy_name), num_sec=180, delay=5)
        except Exception as e:
            self.logger.error(
                "Could not mark {} as a OpenStack template! ({})".format(instance_name, str(e)))
            instance.update(original_name)  # Clean up after ourselves
            raise

    def rename_vm(self, instance_name, new_name):
        instance = self._find_instance_by_name(instance_name)
        try:
            instance.update(new_name)
        except Exception as e:
            self.logger.exception(e)
            return instance_name
        else:
            return new_name

    def delete_template(self, template_name):
        template = self._find_template_by_name(template_name)
        template.delete()
        wait_for(lambda: not self.does_template_exist(template_name), num_sec=120, delay=10)

    def stack_exist(self, stack_name):
        stack = self.stackapi.stacks.get(stack_name)
        if stack:
            return True
        else:
            return False

    def delete_stack(self, stack_name):
        """Deletes stack

        Args:
        stack_name: Unique name of stack
        """

        self.logger.info(" Terminating RHOS stack %s" % stack_name)
        try:
            self.stackapi.stacks.delete(stack_name)
            return True
        except ActionTimedOutError:
            return False

    def set_meta_value(self, instance, key, value):
        instance = self._instance_or_name(instance)
        instance.manager.set_meta_item(
            instance, key, value if isinstance(value, basestring) else json.dumps(value))

    def get_meta_value(self, instance, key):
        instance = self._instance_or_name(instance)
        try:
            data = instance.metadata[key]
            try:
                return json.loads(data)
            except ValueError:
                # Support metadata set by others
                return data
        except KeyError:
            raise KeyError('Metadata {} not found in {}'.format(key, instance.name))



# Imports for backward compatility and convenience
# NOQA all the things because
from base import *  # NOQA
from exceptions import *  # NOQA
from ec2 import EC2System  # NOQA
from openstack import OpenstackSystem  # NOQA
from rhevm import RHEVMSystem  # NOQA
from scvmm import SCVMMSystem  # NOQA
from azure import AzureSystem  # NOQA
from virtualcenter import VMWareSystem  # NOQA
from google import GoogleCloudSystem  # NOQA
from kubernetes import Kubernetes  # NOQA
from openshift import Openshift  # NOQA
from hawkular import Hawkular  # NOQA



# coding: utf-8
"""Backend management system classes

Used to communicate with providers without using CFME facilities
"""
from datetime import datetime
import winrm
import json
import urlparse
import os
from cStringIO import StringIO
from contextlib import contextmanager
from textwrap import dedent

from exceptions import VMInstanceNotFound
from lxml import etree
from wait_for import wait_for

from base import MgmtSystemAPIBase


class AzureSystem(MgmtSystemAPIBase):
    """This class is used to connect to Microsoft Azure Portal via PowerShell AzureRM Module
    """
    STATE_RUNNING = "VM running"
    STATE_STOPPED = "VM deallocated"
    STATE_STARTING = "VM starting"
    STATE_SUSPEND = "VM stopped"
    STATE_PAUSED = "Paused"
    STATES_STEADY = {STATE_RUNNING, STATE_PAUSED}
    STATES_STEADY.update(STATE_STOPPED)

    _stats_available = {
        'num_vm': lambda self: len(self.list_vm()),
        'num_template': lambda self: len(self.list_template()),
    }

    def __init__(self, **kwargs):
        super(AzureSystem, self).__init__(kwargs)
        self.host = kwargs["powershell_host"]
        self.provisioning = kwargs['provisioning']
        self.resource_group = kwargs['provisioning']['resource_group']
        self.storage_container = kwargs['provisioning']['storage_container']
        self.username = kwargs["username"]
        self.password = kwargs["password"]
        self.ui_username = kwargs["ui_username"]
        self.ui_password = kwargs["ui_password"]
        self.ps_username = kwargs["powershell_username"]
        self.ps_password = kwargs["powershell_password"]
        self.storage_account = kwargs["storage_account"]
        self.storage_key = kwargs["storage_key"]
        self.subscription_id = kwargs["subscription_id"]
        self.tenant_id = kwargs["tenant_id"]
        self.region = kwargs["provisioning"]["region_api"]
        self.api = winrm.Session(self.host, auth=(self.ps_username, self.ps_password))

    @property
    def pre_script(self):
        """Script that ensures we can access the Azure Portal Resource Manager.

        SCVMM2 will be used as the PowerShell Host for accessing Azure API
        but any windows machine or VM with AzureRm Powershell module can be used.
        """
        return dedent("""
        $myazurename = "{}"
        $myazurepwd = ConvertTo-SecureString "{}" -AsPlainText -Force
        $azcreds = New-Object System.Management.Automation.PSCredential ($myazurename, $myazurepwd)
        Login-AzureRMAccount -Credential $azcreds
        Get-AzureRmSubscription -SubscriptionId \"{}\" -TenantId \"{}\" | Select-AzureRmSubscription
        """.format(self.ui_username, self.ui_password, self.subscription_id, self.tenant_id))

    def run_script(self, script, ignore_error=False, tries=3):
        """Wrapper for running powershell scripts. Ensures the ``pre_script`` is loaded."""
        script = dedent(script)
        while tries > 0:
            if '\n' not in script:
                # Save space when no newline in the script
                self.logger.info(" Running PowerShell script: {}".format(script))
            else:
                self.logger.info(" Running PowerShell script:\n{}".format(script))
            result = self.api.run_ps("{}\n\n{}".format(self.pre_script, script))
            if result.status_code == 0:
                self.logger.info("run_script Script Complete")
                return result.std_out.strip()
            self.logger.error("Script returned {}!: {}".format(result.status_code, result.std_err))
            tries -= 1
        if result.status_code != 0 and not ignore_error:
            raise self.PowerShellScriptError("Script returned {}!: {}"
                                         .format(result.status_code, result.std_err))

    def start_vm(self, vm_name, resource_group=None):
        if self.is_vm_stopped(vm_name, resource_group or self.resource_group):
            self.logger.info("Attempting to Start Azure VM {}".format(vm_name))
            self.run_script(
                "Start-AzureRmVm -ResourceGroup \"{}\" -Name \"{}\""
                .format(resource_group or self.resource_group, vm_name).strip())
            self.wait_vm_running(vm_name, resource_group)
        else:
            self.logger.info("Azure VM {} is already running".format(vm_name))

    def restart_vm(self, vm_name, resource_group=None):
        self.logger.info("Attempting to Restart Azure VM {}".format(vm_name))
        self.run_script(
            "Restart-AzureRmVm -ResourceGroup \"{}\" -Name \"{}\""
            .format(resource_group or self.resource_group, vm_name).strip())

    def stop_vm(self, vm_name, resource_group=None, shutdown=False):
        if self.is_vm_running(vm_name, resource_group or self.resource_group):
            self.logger.info("Attempting to Stop Azure VM {}".format(vm_name))
            self.run_script(
                "Stop-AzureRmVm -ResourceGroup \"{}\" -Name \"{}\" -Force"
                .format(resource_group or self.resource_group, vm_name).strip())
            self.wait_vm_stopped(vm_name, resource_group)
        else:
            self.logger.info("Azure VM {} is already stopped".format(vm_name))

    def suspend_vm(self, vm_name, resource_group=None):
        if self.is_vm_running(vm_name, resource_group or self.resource_group):
            self.logger.info("Attempting to Suspend Azure VM {}".format(vm_name))
            self.run_script(
                "Stop-AzureRmVm -ResourceGroup \"{}\" -Name \"{}\" -Force -StayProvisioned"
                .format(resource_group or self.resource_group, vm_name).strip())
            self.wait_vm_suspended(vm_name, resource_group)
        else:
            self.logger.info("Azure VM {} is already suspended or stopped".format(vm_name))

    def wait_vm_running(self, vm_name, resource_group=None, num_sec=300):
        wait_for(
            lambda: self.is_vm_running(vm_name, resource_group or self.resource_group),
            message="Waiting for Azure VM {} to be running.".format(vm_name),
            num_sec=num_sec)

    def wait_vm_steady(self, vm_name, resource_group=None, num_sec=300):
        self.logger.info("All states are steady in Azure. {}".format(vm_name))
        return True

    def wait_vm_stopped(self, vm_name, resource_group=None, num_sec=300):
        wait_for(
            lambda: self.is_vm_stopped(vm_name, resource_group or self.resource_group),
            message="Waiting for Azure VM {} to be stopped.".format(vm_name),
            num_sec=num_sec)

    def wait_vm_suspended(self, vm_name, resource_group=None, num_sec=300):
        wait_for(
            lambda: self.is_vm_suspended(vm_name, resource_group or self.resource_group),
            message="Waiting for Azure VM {} to be suspended.".format(vm_name),
            num_sec=num_sec)

    def create_vm(self, vm_name, resource_group=None):
        raise NotImplementedError('NIE - create_vm not implemented.')

    def delete_vm(self, vm_name, resource_group=None):
        self.logger.info("Begin delete_vm {}".format(vm_name))
        vhd_endpoint = self.get_vm_vhd(vm_name, resource_group or self.resource_group)
        self.run_script(
            """
            Invoke-Command -scriptblock {{
            Remove-AzureRmVM -ResourceGroupName \"{rg}\" -Name \"{vm}\" -Force
            Remove-AzureRmNetworkInterface -Name \"{vm}\" -ResourceGroupName \"{rg}\" -Force
            Remove-AzureRmPublicIpAddress -Name \"{vm}\" -ResourceGroupName \"{rg}\" -Force
            }}
            """.format(rg=resource_group or self.resource_group, vm=vm_name), True)

        vhd_name = os.path.split(urlparse.urlparse(vhd_endpoint).path)[1]
        self.remove_blob_image(vhd_name)

    def list_vm(self):
        self.logger.info("Attempting to List Azure VMs")
        azure_data = self.run_script("Get-AzureRmVm | convertto-xml -as String")
        data = self.clean_azure_xml(azure_data)
        nameList = etree.parse(StringIO(data)).getroot().xpath(
            "./Object/Property[@Name='Name']/text()")
        return nameList

    def capture_vm(self, vm_name, resource_group, container, image_name):
        self.logger.info("Attempting to Capture Azure VM {}".format(vm_name))
        self.stop_vm(vm_name, resource_group)

        self.logger.info("Generalizing passed VM {}".format(vm_name))
        self.run_script('Set-AzureRmVM -ResourceGroupName "{}" '
                        '-Name "{}" -Generalized'.format(resource_group, vm_name))
        self.logger.info("The VM {} is generalized".format(vm_name))

        self.run_script('Save-AzureRmVMImage -ResourceGroupName "{}" '
                        '-VMName "{}" -DestinationContainerName "{}" '
                        '-VHDNamePrefix "{}"'.format(resource_group, vm_name, container,
                                                     image_name))
        # save..image always puts image to system
        # https://github.com/Azure/azure-powershell/issues/2714
        if len([image for image in self.list_blob_images('system') if image_name in image]) > 0:
            self.logger.info("Azure VM {} is captured".format(vm_name))
        else:
            raise RuntimeError("image {} isn't found in container {}".format(image_name, container))

    def list_stack(self, resource_group=None):
        self.logger.info("Attempting to List Azure Orchestration Deployment Stacks")
        azure_data = self.run_script(
            """
            Invoke-Command -scriptblock {{
            Get-AzureRmResourceGroupDeployment -ResourceGroupName \"{rg}\" |
            convertto-xml -as String;
            }}
            """.format(rg=resource_group or self.resource_group))
        data = self.clean_azure_xml(azure_data)
        name_list = etree.parse(StringIO(data)).getroot().xpath(
            "./Object/Property[@Name='DeploymentName']/text()")
        return name_list

    def list_template(self):
        self.logger.info("Attempting to List Azure VHDs in templates directory")
        azure_data = self.run_script(
            """
            Invoke-Command -scriptblock {{
            $myStorage = New-AzureStorageContext -StorageAccountName \"{}\" |
            -StorageAccountKey '\"{}\"';
            Get-AzureStorageBlob -Container templates -Context $myStorage | Select Name;
            }}
            """.format(self.storage_account, self.storage_key))
        lines = iter(azure_data.splitlines())
        templates = []
        for line in lines:
            if ".vhd" in line:
                vhd = line.split(" ")
                templates.append(str(vhd[2])[:-4])
        templates.append(str('Microsoft.Compute/Images/templates/tmpl-osDisk'))
        return templates

    def list_flavor(self):
        raise NotImplementedError('list_flavor not implemented.')

    def list_network(self):
        raise NotImplementedError('list_network not implemented.')

    def vm_creation_time(self, vm_name, resource_group=None):
        # There is no such parameter as vm creation time.  Using VHD date instead.
        self.logger.info("Attempting to Retrieve Azure VM Modification Time {}".format(vm_name))
        vm_vhd = self.get_vm_vhd(vm_name, resource_group or self.resource_group)
        vhd_name = os.path.split(urlparse.urlparse(vm_vhd).path)[1]
        data = self.run_script(
            """
            Invoke-Command -scriptblock {{
            $storageContext = New-AzureStorageContext -StorageAccountName \"{storage_account}\" `
                            -StorageAccountKey \"{storage_key}\"
            Get-AzureStorageBlob -Name \"{storage_container}\" `
                            -Context $storageContext -Blob \"{vhd_blob}\" | convertto-xml -as String
            }}
            """.format(storage_account=self.storage_account,
                       storage_container=self.storage_container,
                       storage_key=self.storage_key,
                       vhd_blob=vhd_name), True)
        vhd_last_modified = etree.parse(StringIO(self.clean_azure_xml(data))).getroot().xpath(
            "./Object/Property[@Name='LastModified']/text()")
        create_time = datetime.strptime(str(vhd_last_modified), '[\'%m/%d/%Y %H:%M:%S %p +00:00\']')
        self.logger.info("VM last edit time based on vhd =  {}".format(str(create_time)))
        return create_time

    def create_netsec_group(self, group_name, resource_group):
        self.logger.info("Attempting to Create New Azure Security Group {}".format(group_name))
        self.run_script(
            'New-AzureRmNetworkSecurityGroup -Location "{}" -Name "{}" '
            '-ResourceGroupName "{}"'.format(self.region, group_name, resource_group))
        self.logger.info("Network Security Group {} is created".format(group_name))

    def remove_netsec_group(self, group_name, resource_group):
        self.logger.info("Attempting to Remove Azure Security Group {}".format(group_name))
        self.run_script(
            'Remove-AzureRmNetworkSecurityGroup -Name "{}" '
            '-ResourceGroupName "{}" -Force'.format(group_name, resource_group))
        self.logger.info("Network Security Group {} is removed".format(group_name))

    def info(self, vm_name):
        pass

    def disconnect(self):
        pass

    def clean_azure_xml(self, azure_xml_data):
        # Azure prepends a non-XML header to returned xml.  This strips that header.
        sep = '<?xml'
        clean_xml = sep + azure_xml_data.split(sep, 1)[1]
        return clean_xml

    def vm_status(self, vm_name, resource_group=None):
        self.logger.info("Attempting to Retrieve Azure VM Status {}".format(vm_name))
        azure_data = self.run_script(
            "Get-AzureRmVm -ResourceGroup \"{}\" -Name \"{}\" -Status | convertto-xml -as String"
            .format(resource_group or self.resource_group, vm_name))
        data = self.clean_azure_xml(azure_data)
        statusValue = json.loads(etree.parse(StringIO(data)).getroot().xpath(
            "./Object/Property[@Name='StatusesText']/text()")[0])
        # If script runs completely but the result isn't the one we need - better to show Azure
        # message
        if statusValue[0]['DisplayStatus'] == 'Provisioning failed':
            raise VMInstanceNotFound(statusValue[0]['Message'])
        powerStatus = statusValue[1]
        powerDisplayStatus = powerStatus['DisplayStatus']
        self.logger.info("Returned Status was {}".format(powerDisplayStatus))
        return powerDisplayStatus

    def is_vm_running(self, vm_name, resource_group=None):
        if self.vm_status(vm_name, resource_group or self.resource_group) == self.STATE_RUNNING:
            self.logger.info("According to Azure, the VM \"{}\" is running".format(vm_name))
            return True
        else:
            return False

    def is_vm_stopped(self, vm_name, resource_group=None):
        if self.vm_status(vm_name, resource_group or self.resource_group) == self.STATE_STOPPED:
            self.logger.info("According to Azure, the VM \"{}\" is stopped".format(vm_name))
            return True
        else:
            return False

    def is_vm_starting(self, vm_name, resource_group=None):
        if self.vm_status(vm_name, resource_group or self.resource_group) == self.STATE_STARTING:
            self.logger.info("According to Azure, the VM \"{}\" is starting".format(vm_name))
            return True
        else:
            return False

    def is_vm_suspended(self, vm_name, resource_group=None):
        if self.vm_status(vm_name, resource_group or self.resource_group) == self.STATE_SUSPEND:
            self.logger.info("According to Azure, the VM \"{}\" is suspended".format(vm_name))
            return True
        else:
            return False

    def in_steady_state(self, vm_name, resource_group=None):
        return self.vm_status(vm_name, resource_group or self.resource_group) in self.STATES_STEADY

    def clone_vm(self, source_name, vm_name):
        """It wants exact host and placement (c:/asdf/ghjk) :("""
        raise NotImplementedError('NIE - clone_vm not implemented.')

    def does_vm_exist(self, vm_name):
        return vm_name in self.list_vm()

    def stack_exist(self, stack_name):
        return stack_name in self.list_stack()

    def delete_stack(self, stack_name, resource_group=None):
        self.logger.info("Removes a Deployment Stack resource created with Orchestration")
        self.run_script(
            """
            Invoke-Command -scriptblock {{
            Remove-AzureRmResourceGroupDeployment -ResourceGroupName \"{rg}\" `
            -DeploymentName \"{stack}\" -Force
            }}
            """.format(rg=resource_group or self.resource_group, stack=stack_name))
        return True

    def deploy_template(self, template, vm_name=None, **vm_settings):
        self.copy_blob_image(template, vm_name, vm_settings['storage_account'],
            vm_settings['template_container'], vm_settings['storage_container'])
        self.run_script(
            """
            Invoke-Command -scriptblock {{
            $StorageAccount = Get-AzureRmStorageAccount -ResourceGroupName \"{resource_group}\" `
                 -Name \"{storage_account}\"
            $NetworkSecurityGroupID = Get-AzureRmNetworkSecurityGroup -Name \"{network_nsg}\" `
                -ResourceGroupName \"{resource_group}\"
            $PIp = New-AzureRmPublicIpAddress -Name \"{vm_name}\" -ResourceGroupName `
                \"{resource_group}\" -Location \"{region}\" -AllocationMethod Dynamic -Force
            $SubnetConfig = New-AzureRmVirtualNetworkSubnetConfig -Name "default" `
                -AddressPrefix \"{subnet_range}\"
            $VNet = New-AzureRmVirtualNetwork -Name \"{virtual_net}\" -ResourceGroupName `
                \"{resource_group}\" -Location \"{region}\" -AddressPrefix \"{address_space}\" `
                -Subnet $SubnetConfig -Force
            $Interface = New-AzureRmNetworkInterface -Name \"{vm_name}\" -ResourceGroupName `
                \"{resource_group}\" -Location \"{region}\" -SubnetId $VNet.Subnets[0].Id `
                -PublicIpAddressId $PIp.Id -Force
            $VirtualMachine = New-AzureRmVMConfig -VMName \"{vm_name}\" -VMSize \"{vm_size}\"
            $VirtualMachine = Add-AzureRmVMNetworkInterface -VM $VirtualMachine -Id $Interface.Id
            $OSDiskUri = $StorageAccount.PrimaryEndpoints.Blob.ToString() + \"{storage_cont}\" `
                + "/" + \"{vm_name}.vhd\"
            $VirtualMachine = Set-AzureRmVMOSDisk -VM $VirtualMachine -Name \"{vm_name}\" `
                -VhdUri $OSDiskUri -CreateOption attach -Linux
            New-AzureRmVM -ResourceGroupName \"{resource_group}\" -Location \"{region}\" `
                -VM $VirtualMachine
            }}
            """.format(source_name=template.split("/")[-1],
                       vm_name=vm_name,
                       resource_group=vm_settings['resource_group'],
                       virtual_net=vm_settings['virtual_net'],
                       address_space=vm_settings['address_space'],
                       subnet_range=vm_settings['subnet_range'],
                       network_nsg=vm_settings['network_nsg'],
                       region=vm_settings['region_api'],
                       vm_size=vm_settings['vm_size'],
                       av_set=vm_settings['av_set'],
                       storage_account=vm_settings['storage_account'],
                       storage_cont=vm_settings['storage_container']))
        self.wait_vm_running(vm_name, vm_settings['resource_group'])

    def copy_blob_image(self,
                        template,
                        vm_name,
                        storage_account,
                        template_container,
                        storage_container):
        self.run_script(
            """
            Invoke-Command -scriptblock {{
            $sourceContext = New-AzureStorageContext -StorageAccountName \"{storage_account}\" `
                            -StorageAccountKey \"{storage_key}\"
            $destContext = New-AzureStorageContext -StorageAccountName \"{storage_account}\" `
                            -StorageAccountKey \"{storage_key}\"
            $blobCopy = Start-AzureStorageBlobCopy -DestContainer \"{storage_container}\" `
                        -DestContext $destContext -DestBlob \"{vm_name}.vhd\" `
                        -SrcBlob \"{source_name}.vhd\" `
                        -Context $sourceContext -SrcContainer \"{template_container}\"
            }}
            """.format(source_name=template.split("/")[-1],
                       vm_name=vm_name,
                       storage_account=storage_account,
                       storage_key=self.storage_key,
                       template_container=template_container,
                       storage_container=storage_container))

    def remove_blob_image(self, vm_vhd, container=None):
        if not container:
            container = self.storage_container

        self.run_script(
            """
            Invoke-Command -scriptblock {{
            $sourceContext = New-AzureStorageContext -StorageAccountName \"{storage_account}\" `
                            -StorageAccountKey \"{storage_key}\"
            $blobRemove = Remove-AzureStorageBlob -Blob \"{vhd_name}\" `
                        -Context $sourceContext -Container \"{container}\"
            }}
            """.format(vhd_name=vm_vhd,
                       storage_account=self.storage_account,
                       storage_key=self.storage_key,
                       container=container), True)

    def list_blob_images(self, container):
        azure_data = self.run_script(
            'New-AzureStorageContext -StorageAccountName "{acc_name}"'
            ' -StorageAccountKey "{acc_key}"|Get-AzureStorageBlob -Container "{container}"|'
            'Select Name|convertto-xml -as String'.format(acc_name=self.storage_account,
                                                          acc_key=self.storage_key,
                                                          container=container))
        data = self.clean_azure_xml(azure_data)
        names = etree.parse(StringIO(data)).getroot().xpath(
            "./Object/Property[@Name='Name']/text()")
        return names

    def remove_diags_container(self):
        self.run_script(
            """
            Invoke-Command -scriptblock {{
            $storeContext = New-AzureStorageContext -StorageAccountName \"{storage_account}\" `
                            -StorageAccountKey \"{storage_key}\"
            $storageContainer = Get-AzureStorageContainer -Context $storeContext
            foreach ($container in $storageContainer) {{
                if ($container.name -like 'bootdiagnostics-test*'){{
                    Remove-AzureStorageContainer -Name $container.name -Context $storeContext -Force
                }}
            }}
            }}
            """.format(storage_account=self.storage_account,
                       storage_key=self.storage_key), True)

    @contextmanager
    def with_vm(self, *args, **kwargs):
        """Context manager for better cleanup"""
        name = self.deploy_template(*args, **kwargs)
        yield name
        self.delete_vm(name)

    def current_ip_address(self, vm_name, resource_group=None):
        # Returns first active IPv4 IpAddress only
        azure_data = self.run_script(
            "Get-AzureRmPublicIpAddress -ResourceGroup \"{}\" -Name \"{}\" |"
            "convertto-xml -as String".format(resource_group or self.resource_group, vm_name))
        data = self.clean_azure_xml(azure_data)
        return etree.parse(StringIO(data)).getroot().xpath(
            "./Object/Property[@Name='IpAddress']/text()")

    def get_ip_address(self, vm_name, resource_group=None, **kwargs):
        current_ip_address = self.current_ip_address(vm_name, resource_group or self.resource_group)
        return current_ip_address

    def get_vm_vhd(self, vm_name, resource_group=None):
        self.logger.info("get_vm_vhd - Attempting to Retrieve Azure VM VHD {}".format(vm_name))
        azure_data = self.run_script(
            "Get-AzureRmVm -ResourceGroup \"{}\" -Name \"{}\" | convertto-xml -as String"
            .format(resource_group or self.resource_group, vm_name))
        data = self.clean_azure_xml(azure_data)
        vhd_value = json.loads(etree.parse(StringIO(data)).getroot().xpath(
            "./Object/Property[@Name='StorageProfileText']/text()")[0])
        vhd_endpoint = vhd_value['OSDisk']['VirtualHardDisk']['Uri']
        self.logger.info("Returned Disk Endpoint was {}".format(vhd_endpoint))
        return vhd_endpoint

    def get_network_interface(self, vm_name, resource_group=None):
        self.logger.info("Attempting to Retrieve Azure VM Network Interface {}".format(vm_name))
        azure_data = self.run_script(
            "Get-AzureRmVm -ResourceGroup \"{}\" -Name \"{}\" | convertto-xml -as String"
            .format(resource_group or self.resource_group, vm_name))
        data = self.clean_azure_xml(azure_data)
        status_value = json.loads(etree.parse(StringIO(data)).getroot().xpath(
            "./Object/Property[@Name='NetworkProfileText']/text()")[0])
        nic_uri = status_value['NetworkInterfaces'][0]['ReferenceUri']
        self.logger.info("Returned URI was {}".format(nic_uri))
        return os.path.split(urlparse.urlparse(nic_uri).path)[1]

    def remove_host_from_cluster(self, hostname):
        """I did not notice any scriptlet that lets you do this."""

    def disconnect_dvd_drives(self, vm_name):
        raise NotImplementedError('disconnect_dvd_drives not implemented.')

    def data(self, vm_name, resource_group=None):
        raise NotImplementedError('data not implemented.')

    ##
    # Classes and functions used to access detailed SCVMM Data
    @staticmethod
    def parse_data(t, data):
        if data is None:
            return None
        elif t == "System.Boolean":
            return data.lower().strip() == "true"
        elif t.startswith("System.Int"):
            return int(data)
        elif t == "System.String" and data.lower().strip() == "none":
            return None

    class AzureDataHolderDict(object):
        def __init__(self, data):
            for prop in data.xpath("./Property"):
                name = prop.attrib["Name"]
                t = prop.attrib["Type"]
                children = prop.getchildren()
                if children:
                    if prop.xpath("./Property[@Name]"):
                        self.__dict__[name] = AzureSystem.AzureDataHolderDict(prop)
                    else:
                        self.__dict__[name] = AzureSystem.AzureDataHolderList(prop)
                else:
                    data = prop.text
                    result = AzureSystem.parse_data(t, prop.text)
                    self.__dict__[name] = result

        def __repr__(self):
            return repr(self.__dict__)

    class AzureDataHolderList(list):
        def __init__(self, data):
            super(AzureSystem.AzureDataHolderList, self).__init__()
            for prop in data.xpath("./Property"):
                t = prop.attrib["Type"]
                data = prop.text
                result = AzureSystem.parse_data(t, prop.text)
                self.append(result)

    class PowerShellScriptError(Exception):
        pass



# coding: utf-8
"""Backend management system classes

Used to communicate with providers without using CFME facilities
"""
try:
    # In Fedora 22, we see SSL errors when connecting to vSphere, this prevents the error.
    import ssl
    ssl._create_default_https_context = ssl._create_unverified_context
except AttributeError:
    pass

import re
from datetime import datetime, timedelta
from distutils.version import LooseVersion
from functools import partial
import operator

import time
from psphere import managedobjects as mobs
from psphere.client import Client
from psphere.errors import ObjectNotFoundError
from suds import WebFault
from wait_for import wait_for, TimedOutError

from base import MgmtSystemAPIBase, VMInfo
from exceptions import VMInstanceNotCloned, VMInstanceNotSuspended, VMNotFoundViaIP, HostNotRemoved


class _PsphereClient(Client):

    def __init__(self, *args, **kwargs):
        self._cached_retry = dict()
        self.logger = kwargs['logger']
        kwargs.pop('logger')
        super(_PsphereClient, self).__init__(*args, **kwargs)

    def get_search_filter_spec(self, *args, **kwargs):
        # A datastore traversal spec is missing from this method in psphere.
        # psav has opened a PR to add it, but until it gets merged we'll need to come behind
        # psphere and add it in just like his PR does
        # https://github.com/jkinred/psphere/pull/18/files
        pfs = super(_PsphereClient, self).get_search_filter_spec(*args, **kwargs)
        select_sets = pfs.objectSet[0].selectSet
        missing_ss = 'datacenter_datastore_traversal_spec'
        ss_names = [ss.name for ss in select_sets]

        if missing_ss not in ss_names:
            self.logger.trace('Injecting %s into psphere search filter spec', missing_ss)
            # pull out the folder traversal spec traversal specs
            fts_ts = pfs.objectSet[0].selectSet[0]
            # and get the select set from the traversal spec
            fts_ss = fts_ts.selectSet[0]

            # add ds selection spec to folder traversal spec
            dsss = self.create('SelectionSpec', name=missing_ss)
            fts_ts.selectSet.append(dsss)

            # add ds traversal spec to search filter object set select spec
            dsts = self.create('TraversalSpec')
            dsts.name = 'datacenter_datastore_traversal_spec'
            dsts.type = 'Datacenter'
            dsts.path = 'datastoreFolder'
            dsts.selectSet = [fts_ss]
            select_sets.append(dsts)
        else:
            self.logger.warning(
                "%s already in psphere search filer spec, not adding it", missing_ss)

        return pfs

    def __getattribute__(self, attr):
        # fetch the attribute using parent class to avoid recursion
        res = super(_PsphereClient, self).__getattribute__(attr)
        # any callable (except 'login') is protected against unexpected logout
        if callable(res) and attr not in ('login', '_login_retry_wrapper'):
            if attr not in self._cached_retry:
                self._cached_retry[attr] = self._login_retry_wrapper(res)
            return self._cached_retry[attr]
        # don't mess with non-callables - just return them
        return res

    def _login_retry_wrapper(self, o):
        # tries to log in on failure
        def f(*args, **kwargs):
            try:
                return o(*args, **kwargs)
            except ObjectNotFoundError:
                try:
                    self.logout()
                except WebFault:
                    # Server raíses the following when we try to logout with an old session
                    # WebFault: Server raised fault: 'The session is not authenticated.'
                    pass
                self.logger.debug("{} disconnected (psphere api); logging back in and trying again"
                    .format(self.server))
                self.login()
                return o(*args, **kwargs)
        return f


class VMWareSystem(MgmtSystemAPIBase):
    """Client to Vsphere API

    Args:
        hostname: The hostname of the system.
        username: The username to connect with.
        password: The password to connect with.

    See also:

        vSphere Management SDK API docs
        https://developercenter.vmware.com/web/dp/doc/preview?id=155

    """
    _api = None

    _stats_available = {
        'num_vm': lambda self: len(self.list_vm()),
        'num_host': lambda self: len(self.list_host()),
        'num_cluster': lambda self: len(self.list_cluster()),
        'num_template': lambda self: len(self.list_template()),
        'num_datastore': lambda self: len(self.list_datastore()),
    }
    POWERED_ON = 'poweredOn'
    POWERED_OFF = 'poweredOff'
    SUSPENDED = 'suspended'

    def __init__(self, hostname, username, password, **kwargs):
        super(VMWareSystem, self).__init__(kwargs)
        self.hostname = hostname
        self.username = username
        self.password = password
        self._api = None
        self._vm_cache = {}
        self.kwargs = kwargs

    @property
    def api(self):
        if not self._api:
            self._api = _PsphereClient(self.hostname, self.username, self.password,
                logger=self.logger)
        return self._api

    @property
    def version(self):
        return LooseVersion(self.api.si.content.about.version)

    @property
    def default_resource_pool(self):
        return self.kwargs.get("default_resource_pool", None)

    def _get_vm(self, vm_name, force=False):
        """ Returns a vm from the VI object.

        Args:
            vm_name: The name of the VM.
            force: Ignore the cache when updating

        Returns: a psphere object.
        """
        if vm_name not in self._vm_cache or force:
            self._vm_cache[vm_name] = mobs.VirtualMachine.get(self.api, name=vm_name)
        else:
            self._vm_cache[vm_name].update()

        return self._vm_cache[vm_name]

    def _get_resource_pool(self, resource_pool_name=None):
        """ Returns a resource pool MOR for a specified name.

        Args:
            resource_pool_name: The name of the resource pool. If None, first one will be picked.
        Returns: The MOR of the resource pool.
        """
        if resource_pool_name is not None:
            return mobs.ResourcePool.get(self.api, name=resource_pool_name)
        elif self.default_resource_pool is not None:
            return mobs.ResourcePool.get(self.api, name=self.default_resource_pool)
        else:
            return mobs.ResourcePool.all(self.api)[0]

    @staticmethod
    def _task_wait(task):
        task.update()
        if task.info.state not in ['queued', 'running', None]:
            return task.info.state

    def does_vm_exist(self, name):
        """ Checks if a vm exists or not.

        Args:
            name: The name of the requested vm.
        Returns: A boolean, ``True`` if the vm exists, ``False`` if not.
        """
        try:
            self._get_vm(name)
            return True
        except Exception:
            return False

    def current_ip_address(self, vm_name):
        ipv4_re = r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}'
        try:
            vm = self._get_vm(vm_name)
            ip_address = vm.summary.guest.ipAddress
            if not re.match(ipv4_re, ip_address) or ip_address == '127.0.0.1':
                ip_address = None
            return ip_address
        except (AttributeError, TypeError):
            # AttributeError: vm doesn't have an ip address yet
            # TypeError: ip address wasn't a string
            return None

    def get_ip_address(self, vm_name, timeout=600):
        """ Returns the first IP address for the selected VM.

        Args:
            vm_name: The name of the vm to obtain the IP for.
            timeout: The IP address wait timeout.
        Returns: A string containing the first found IP that isn't the loopback device.
        """
        try:
            ip_address, tc = wait_for(lambda: self.current_ip_address(vm_name),
                fail_condition=None, delay=5, num_sec=timeout,
                message="get_ip_address from vsphere")
        except TimedOutError:
            ip_address = None
        return ip_address

    def _get_list_vms(self, get_template=False, inaccessible=False):
        """ Obtains a list of all VMs on the system.

        Optional flag to obtain template names too.

        Args:
            get_template: A boolean describing if it should return template names also.
        Returns: A list of VMs.
        """
        # Use some psphere internals to get vm propsets back directly with requested properties,
        # so we skip the network overhead of returning full managed objects
        property_spec = self.api.create('PropertySpec')
        property_spec.all = False
        property_spec.pathSet = ['name', 'config.template', 'config.uuid',
            'runtime.connectionState']
        property_spec.type = 'VirtualMachine'
        pfs = self.api.get_search_filter_spec(self.api.si.content.rootFolder, property_spec)
        object_contents = self.api.si.content.propertyCollector.RetrieveProperties(specSet=[pfs])

        # Ensure get_template is either True or False to match the config.template property
        get_template = bool(get_template)

        # Select the vms or templates based on get_template and the returned properties
        obj_list = []
        for object_content in object_contents:
            # Nested property lookups work, but the attr lookup on the
            # vm object still triggers a request even though the vm
            # object already "knows" the answer in its cached object
            # content. So we just pull the value straight out of the cache.
            vm_props = {p.name: p.val for p in object_content.propSet}
            if vm_props.get('config.template') == get_template:
                if (vm_props.get('runtime.connectionState') == "inaccessible"
                        and inaccessible) or vm_props.get(
                            'runtime.connectionState') != "inaccessible":
                    obj_list.append(vm_props['name'])
        return obj_list

    def all_vms(self):
        property_spec = self.api.create('PropertySpec')
        property_spec.all = False
        property_spec.pathSet = ['name', 'config.template']
        property_spec.type = 'VirtualMachine'
        pfs = self.api.get_search_filter_spec(self.api.si.content.rootFolder, property_spec)
        object_contents = self.api.si.content.propertyCollector.RetrieveProperties(specSet=[pfs])
        result = []
        for vm in object_contents:
            vm_props = {p.name: p.val for p in vm.propSet}
            if vm_props.get('config.template'):
                continue
            try:
                ip = str(vm.obj.summary.guest.ipAddress)
            except AttributeError:
                ip = None
            try:
                uuid = str(vm.obj.summary.config.uuid)
            except AttributeError:
                uuid = None
            result.append(
                VMInfo(
                    uuid,
                    str(vm.obj.summary.config.name),
                    str(vm.obj.summary.runtime.powerState),
                    ip,
                )
            )
        return result

    def get_vm_guid(self, vm_name):
        vm = self._get_vm(vm_name)
        try:
            return str(vm.summary.config.uuid)
        except AttributeError:
            return None

    def get_vm_name_from_ip(self, ip):
        """ Gets the name of a vm from its IP.

        Args:
            ip: The ip address of the vm.
        Returns: The vm name for the corresponding IP."""
        vms = self.api.si.content.searchIndex.FindAllByIp(ip=ip, vmSearch=True)
        # As vsphere remembers the last IP a vm had, when we search we get all
        # of them. Consequently we need to store them all in a dict and then sort
        # them to find out which one has the latest boot time. I am going out on
        # a limb and saying that searching for several vms and querying each object
        # is quicker than finding all machines and recording the bootTime and ip address
        # of each, before iterating through all of them to weed out the ones we care
        # about, but I could be wrong.
        boot_times = {}
        for vm in vms:
            if vm.name not in boot_times:
                boot_times[vm.name] = datetime.fromtimestamp(0)
                try:
                    boot_times[vm.name] = vm.summary.runtime.bootTime
                except:
                    pass
        if boot_times:
            newest_boot_time = sorted(boot_times.items(), key=operator.itemgetter(1),
                                      reverse=True)[0]
            return newest_boot_time[0]
        else:
            raise VMNotFoundViaIP('The requested IP is not known as a VM')

    def start_vm(self, vm_name):
        self.wait_vm_steady(vm_name)
        if self.is_vm_running(vm_name):
            self.logger.info(" vSphere VM %s is already running" % vm_name)
            return True
        else:
            self.logger.info(" Starting vSphere VM %s" % vm_name)
            vm = self._get_vm(vm_name)
            vm.PowerOnVM_Task()
            self.wait_vm_running(vm_name)
            return True

    def stop_vm(self, vm_name):
        self.wait_vm_steady(vm_name)
        if self.is_vm_stopped(vm_name):
            self.logger.info(" vSphere VM %s is already stopped" % vm_name)
            return True
        else:
            self.logger.info(" Stopping vSphere VM %s" % vm_name)
            vm = self._get_vm(vm_name)
            if self.is_vm_suspended(vm_name):
                self.logger.info(
                    " Resuming suspended VM %s before stopping." % vm_name
                )
                vm.PowerOnVM_Task()
                self.wait_vm_running(vm_name)
            vm.PowerOffVM_Task()
            self.wait_vm_stopped(vm_name)
            return True

    def delete_vm(self, vm_name):
        self.wait_vm_steady(vm_name)
        self.logger.info(" Deleting vSphere VM %s" % vm_name)
        vm = self._get_vm(vm_name)
        self.stop_vm(vm_name)

        task = vm.Destroy_Task()
        status, t = wait_for(self._task_wait, [task])
        return status == 'success'

    def is_host_connected(self, host_name):
        host = mobs.HostSystem.get(self.api, name=host_name)
        return True if host.summary.runtime.connectionState == "connected" else False

    def create_vm(self, vm_name):
        raise NotImplementedError('This function has not yet been implemented.')

    def restart_vm(self, vm_name):
        self.logger.info(" Restarting vSphere VM %s" % vm_name)
        return self.stop_vm(vm_name) and self.start_vm(vm_name)

    def list_vm(self, inaccessible=False):
        return self._get_list_vms(inaccessible=inaccessible)

    def list_template(self):
        return self._get_list_vms(get_template=True)

    def list_flavor(self):
        raise NotImplementedError('This function is not supported on this platform.')

    def list_host(self):
        return [str(h.name) for h in mobs.HostSystem.all(self.api)]

    def list_host_datastore_url(self, host_name):
        host = mobs.HostSystem.get(self.api, name=host_name)
        return [str(d.summary.url) for d in host.datastore]

    def list_datastore(self):
        return [str(h.name) for h in mobs.Datastore.all(self.api) if h.host]

    def list_cluster(self):
        return [str(h.name) for h in mobs.ClusterComputeResource.all(self.api)]

    def list_resource_pools(self):
        return [str(h.name) for h in mobs.ResourcePool.all(self.api)]

    def info(self):
        return '%s %s' % (self.api.get_server_type(), self.api.get_api_version())

    def connect(self):
        pass

    def disconnect(self):
        pass

    def vm_status(self, vm_name):
        return self._get_vm(vm_name, force=True).runtime.powerState

    def vm_creation_time(self, vm_name):
        vm = self._get_vm(vm_name)
        vcenter_time_now = self.api.si.CurrentTime()
        vm_uptime = vm.summary.quickStats.uptimeSeconds
        vm_delta = timedelta(seconds=int(vm_uptime))
        return vcenter_time_now - vm_delta

    def get_vm_host_name(self, vm_name):
        vm = self._get_vm(vm_name)
        return str(vm.runtime.host.name)

    def get_vm_datastore_path(self, vm_name, vm_config_datastore):
        vm = self._get_vm(vm_name)
        datastore_url = [str(datastore['url']) for datastore in vm.config.datastoreUrl
                         if datastore['name'] in vm_config_datastore]
        return datastore_url[0]

    def get_vm_config_files_path(self, vm_name):
        vm = self._get_vm(vm_name)
        vmfilespath = vm.config.files.vmPathName
        return str(vmfilespath)

    def in_steady_state(self, vm_name):
        return self.vm_status(vm_name) in {self.POWERED_ON, self.POWERED_OFF, self.SUSPENDED}

    def is_vm_running(self, vm_name):
        return self.vm_status(vm_name) == self.POWERED_ON

    def wait_vm_running(self, vm_name, num_sec=240):
        self.logger.info(" Waiting for vSphere VM %s to change status to ON" % vm_name)
        wait_for(self.is_vm_running, [vm_name], num_sec=num_sec)

    def is_vm_stopped(self, vm_name):
        return self.vm_status(vm_name) == self.POWERED_OFF

    def wait_vm_stopped(self, vm_name, num_sec=240):
        self.logger.info(" Waiting for vSphere VM %s to change status to OFF" % vm_name)
        wait_for(self.is_vm_stopped, [vm_name], num_sec=num_sec)

    def is_vm_suspended(self, vm_name):
        return self.vm_status(vm_name) == self.SUSPENDED

    def wait_vm_suspended(self, vm_name, num_sec=360):
        self.logger.info(" Waiting for vSphere VM %s to change status to SUSPENDED" % vm_name)
        wait_for(self.is_vm_suspended, [vm_name], num_sec=num_sec)

    def suspend_vm(self, vm_name):
        self.wait_vm_steady(vm_name)
        self.logger.info(" Suspending vSphere VM %s" % vm_name)
        vm = self._get_vm(vm_name)
        if self.is_vm_stopped(vm_name):
            raise VMInstanceNotSuspended(vm_name)
        else:
            vm.SuspendVM_Task()
            self.wait_vm_suspended(vm_name)
            return True

    def rename_vm(self, vm_name, new_vm_name):
        vm = self._get_vm(vm_name)
        task = vm.Rename_Task(newName=new_vm_name)
        # Cycle until the new named vm is found
        # That must happen or the error state can come up too
        while not self.does_vm_exist(new_vm_name):
            task.update()
            if task.info.state == "error":
                return vm_name  # Old vm name if error
            time.sleep(0.5)
        else:
            # The newly renamed VM is found
            return new_vm_name

    @staticmethod
    def _progress_log_callback(logger, source, destination, progress):
        logger.info("Provisioning progress {}->{}: {}".format(
            source, destination, str(progress)))

    def _pick_datastore(self, allowed_datastores):
        # Pick a datastore by space
        possible_datastores = [
            ds for ds in mobs.Datastore.all(self.api)
            if ds.name in allowed_datastores and ds.summary.accessible
            and ds.summary.multipleHostAccess and ds.overallStatus != "red"]
        possible_datastores.sort(
            key=lambda ds: float(ds.summary.freeSpace) / float(ds.summary.capacity),
            reverse=True)
        if not possible_datastores:
            raise Exception("No possible datastores!")
        return possible_datastores[0]

    def clone_vm(self, source, destination, resourcepool=None, datastore=None, power_on=True,
                 sparse=False, template=False, provision_timeout=1800, progress_callback=None,
                 allowed_datastores=None, cpu=None, ram=None, **kwargs):
        try:
            if mobs.VirtualMachine.get(self.api, name=destination).name == destination:
                raise Exception("VM already present!")
        except ObjectNotFoundError:
            pass

        if progress_callback is None:
            progress_callback = partial(self._progress_log_callback, self.logger,
                source, destination)

        source_template = mobs.VirtualMachine.get(self.api, name=source)

        vm_clone_spec = self.api.create("VirtualMachineCloneSpec")
        vm_reloc_spec = self.api.create("VirtualMachineRelocateSpec")
        # DATASTORE
        if isinstance(datastore, basestring):
            vm_reloc_spec.datastore = mobs.Datastore.get(self.api, name=datastore)
        elif isinstance(datastore, mobs.Datastore):
            vm_reloc_spec.datastore = datastore
        elif datastore is None:
            if allowed_datastores is not None:
                # Pick a datastore by space
                vm_reloc_spec.datastore = self._pick_datastore(allowed_datastores)
            else:
                # Use the same datastore
                datastores = source_template.datastore
                if isinstance(datastores, (list, tuple)):
                    vm_reloc_spec.datastore = datastores[0]
                else:
                    vm_reloc_spec.datastore = datastores
        else:
            raise NotImplementedError("{} not supported for datastore".format(datastore))
        progress_callback("Picked datastore `{}`".format(vm_reloc_spec.datastore.name))

        # RESOURCE POOL
        if isinstance(resourcepool, mobs.ResourcePool):
            vm_reloc_spec.pool = resourcepool
        else:
            vm_reloc_spec.pool = self._get_resource_pool(resourcepool)
        progress_callback("Picked resource pool `{}`".format(vm_reloc_spec.pool.name))

        vm_reloc_spec.host = None
        if sparse:
            vm_reloc_spec.transform = self.api.create('VirtualMachineRelocateTransformation').sparse
        else:
            vm_reloc_spec.transform = self.api.create('VirtualMachineRelocateTransformation').flat

        vm_clone_spec.powerOn = power_on
        vm_clone_spec.template = template
        vm_clone_spec.location = vm_reloc_spec
        vm_clone_spec.snapshot = None

        if cpu is not None:
            vm_clone_spec.config.numCPUs = int(cpu)
        if ram is not None:
            vm_clone_spec.config.memoryMB = int(ram)

        try:
            folder = source_template.parent.parent.vmParent
        except AttributeError:
            folder = source_template.parent

        progress_callback("Picked folder `{}`".format(folder.name))

        task = source_template.CloneVM_Task(folder=folder, name=destination, spec=vm_clone_spec)

        def _check():
            try:
                progress_callback("{}/{}%".format(task.info.state, task.info.progress))
            except AttributeError:
                pass
            return task.info.state not in {"queued", "running"}

        wait_for(
            _check,
            fail_func=task.update, num_sec=provision_timeout, delay=4
        )
        if task.info.state != 'success':
            self.logger.error('Clone VM failed: {}'.format(task.info.error.localizedMessage))
            raise VMInstanceNotCloned(source)
        else:
            return destination

    def mark_as_template(self, vm_name, **kwargs):
        mobs.VirtualMachine.get(self.api, name=vm_name).MarkAsTemplate()  # Returns None

    def deploy_template(self, template, **kwargs):
        kwargs["power_on"] = kwargs.pop("power_on", True)
        kwargs["template"] = False
        destination = kwargs.pop("vm_name")
        start_timeout = kwargs.pop("timeout", 1800)
        self.clone_vm(template, destination, **kwargs)
        if kwargs["power_on"]:
            self.wait_vm_running(destination, num_sec=start_timeout)
        else:
            self.wait_vm_stopped(destination, num_sec=start_timeout)
        return destination

    def remove_host_from_cluster(self, host_name):
        host = mobs.HostSystem.get(self.api, name=host_name)
        task = host.DisconnectHost_Task()
        status, t = wait_for(self._task_wait, [task])

        if status != 'success':
            raise HostNotRemoved("Host {} not removed: {}".format(
                host_name, task.info.error.localizedMessage))

        task = host.Destroy_Task()
        status, t = wait_for(self._task_wait, [task], fail_condition=None)

        return status == 'success'



# coding: utf-8
"""Backend management system classes

Used to communicate with providers without using CFME facilities
"""

from apiclient.discovery import build
from apiclient.http import MediaFileUpload
from apiclient import errors
from base import MgmtSystemAPIBase, VMInfo
from exceptions import VMInstanceNotFound, ImageNotFoundError, ActionNotSupported, \
    ForwardingRuleNotFound
from json import dumps as json_dumps
from oauth2client.service_account import ServiceAccountCredentials
from wait_for import wait_for
import os
import httplib2
import iso8601
import random
import time

# Retry transport and file IO errors.
RETRYABLE_ERRORS = (httplib2.HttpLib2Error, IOError)
# Number of times to retry failed downloads.
NUM_RETRIES = 5
# Number of bytes to send/receive in each request.
CHUNKSIZE = 2 * 1024 * 1024
# Mimetype to use if one can't be guessed from the file extension.
DEFAULT_MIMETYPE = 'application/octet-stream'

# List of image project which gcr provided from the box. Could be extend in the futute and
# will have impact on total number of templates/images
IMAGE_PROJECTS = ['centos-cloud', 'debian-cloud', 'rhel-cloud', 'suse-cloud', 'ubuntu-os-cloud',
                'windows-cloud', 'opensuse-cloud', 'coreos-cloud', 'google-containers']


class GoogleCloudSystem(MgmtSystemAPIBase):
    """
    Client to Google Cloud Platform API

    """

    _stats_available = {
        'num_vm': lambda self: len(self.all_vms()),
        'num_template': lambda self: len(self.list_image()),
    }

    default_scope = ['https://www.googleapis.com/auth/cloud-platform']
    states = {
        'running': ('RUNNING',),
        'stopped': ('TERMINATED',),
        'starting': ('STAGING'),
        'stopping': ('STOPPING'),
    }

    def __init__(self, project=None, zone=None, file_type=None, **kwargs):
        """
            The last three argumets are optional and required only if you want
            to use json or p12 files.
            By default, we expecting that service_account arg contains service account data.

            Args:
                project: name of the project, so called project_id
                zone: zone of cloud
                service_account: service_account_content

                scope: compute engine, container engine, sqlservice end etc
                file_path: path to json or p12 file
                file_type: p12 or json
                client_email: Require for p12 file

            Returns: A :py:class:`GoogleCloudSystem` object.
        """
        super(GoogleCloudSystem, self).__init__(kwargs)
        self._project = project
        self._zone = zone
        scope = kwargs.get('scope', self.default_scope)

        service_account = kwargs.get('service_account', None)
        if service_account:
            service_account = dict(service_account.items())
            service_account['private_key'] = service_account['private_key'].replace('\\n', '\n')
            credentials = ServiceAccountCredentials.from_json_keyfile_dict(
                service_account, scopes=scope)
        elif file_type == 'json':
            file_path = kwargs.get('file_path', None)
            credentials = ServiceAccountCredentials.from_json_keyfile_name(
                file_path, scopes=scope)
        elif file_type == 'p12':
            file_path = kwargs.get('file_path', None)
            client_email = kwargs.get('client_email', None)
            credentials = ServiceAccountCredentials.from_p12_keyfile(
                client_email, file_path, scopes=scope)
        http_auth = credentials.authorize(httplib2.Http())
        self._compute = build('compute', 'v1', http=http_auth)
        self._storage = build('storage', 'v1', http=http_auth)
        self._instances = self._compute.instances()
        self._forwarding_rules = self._compute.forwardingRules()
        self._buckets = self._storage.buckets()

    def _get_zone_instances(self, zone):
        return self._instances.list(project=self._project, zone=zone).execute()

    def _get_all_buckets(self):
        return self._buckets.list(project=self._project).execute()

    def _get_all_forwarding_rules(self):
        results = []
        results.extend(self._forwarding_rules.list(project=self._project, region=self._zone).
                       execute().get('items', []))
        return results

    def _get_all_images(self):
        images = self._compute.images()
        result = []
        for image_project in IMAGE_PROJECTS:
            result.extend(images.list(project=image_project).execute().get('items', []))
        result.extend(images.list(project=self._project).execute().get('items', []))
        return result

    def get_private_images(self):
        images = self._compute.images()
        return images.list(project=self._project).execute()

    def list_vm(self):
        instances = self.all_vms()
        return [instance.name for instance in instances]

    def list_bucket(self):
        buckets = self._get_all_buckets()
        return [bucket.get('name') for bucket in buckets.get('items', [])]

    def list_forwarding_rules(self):
        rules = self._get_all_forwarding_rules()
        return [forwarding_rule.get('name') for forwarding_rule in rules]

    def list_image(self):
        images = self._get_all_images()
        return [image.get('name') for image in images]

    def _find_instance_by_name(self, instance_name):
        try:
            instance = self._instances.get(
                project=self._project, zone=self._zone, instance=instance_name).execute()
            return instance
        except Exception as e:
            self.logger.error(e)
            self.logger.info("Searching instance {} in all other zones".format(instance_name))
            zones = self._compute.zones().list(project=self._project).execute()
            for zone in zones.get('items', []):
                zone_name = zone.get('name', None)
                for instance in self._get_zone_instances(zone_name).get('items', []):
                    if instance['name'] == instance_name:
                        return instance
            self.logger.error("Instance {} not found in any of the zones".format(instance_name))
            raise VMInstanceNotFound(instance_name)

    def _find_forwarding_rule_by_name(self, forwarding_rule_name):
        try:
            forwarding_rule = self._forwarding_rules.get(
                project=self._project, zone=self._zone,
                forwardingRule=forwarding_rule_name).execute()
            return forwarding_rule
        except Exception:
            raise ForwardingRuleNotFound

    def get_image_by_name(self, image_name):
        try:
            image = self._compute.images().get(project=self._project, image=image_name).execute()
            return image
        except Exception:
            raise ImageNotFoundError(image_name)

    def _nested_operation_wait(self, operation_name, zone=True):
        if not zone:
            result = self._compute.globalOperations().get(
                project=self._project,
                operation=operation_name).execute()
        else:
            result = self._compute.zoneOperations().get(
                project=self._project,
                zone=self._zone,
                operation=operation_name).execute()

        if result['status'] == 'DONE':
            self.logger.info("The operation {} -> DONE".format(operation_name))
            if 'error' in result:
                self.logger.error("Error during {} operation.".format(operation_name))
                self.logger.error("Detailed information about error {}".format(result['error']))
                raise Exception(result['error'])
            return True

        return False

    def create_bucket(self, bucket_name):
        """ Create bucket
        Args:
            bucket_name: Unique name of bucket
        """
        if not self.bucket_exists(bucket_name):
            self._buckets.insert(
                project=self._project, body={"name": "{}".format(bucket_name)}).execute()
            self.logger.info("Bucket {} was created".format(bucket_name))
        else:
            self.logger.info("Bucket {} was not created, exists already".format(bucket_name))

    def create_image(self, image_name, bucket_url, timeout=360):
        """ Create image from file
        Args:
            image_name: Unique name of image
            bucket_url: url to image file in bucket
            timeout: time to wait for operation
        """
        images = self._compute.images()
        data = {
            "name": image_name,
            "rawDisk": {"source": bucket_url}
        }
        operation = images.insert(project=self._project, body=data).execute()
        wait_for(lambda: self._nested_operation_wait(operation['name'], zone=False), delay=0.5,
            num_sec=timeout, message=" Creating image {}".format(image_name))

    def delete_bucket(self, bucket_name):
        """ Delete bucket
        Args:
            bucket_name: Name of bucket
        """
        if self.bucket_exists(bucket_name):
            self._buckets.delete("{}".format(bucket_name)).execute()
            self.logger.info("Bucket {} was deleted".format(bucket_name))
        else:
            self.logger.info("Bucket {} was not deleted, not found".format(bucket_name))

    def bucket_exists(self, bucket_name):
        try:
            self._buckets.get(bucket=bucket_name).execute()
            return True
        except errors.HttpError as error:
            if "Not Found" in error.content:
                self.logger.info("Bucket {} was not found".format(bucket_name))
                return False
            if "Invalid bucket name" in error.content:
                self.logger.info("Incorrect bucket name {} was specified".format(bucket_name))
                return False
            raise error

    def get_file_from_bucket(self, bucket_name, file_name):
        if self.bucket_exists(bucket_name):
            try:
                data = self._storage.objects().get(bucket=bucket_name, object=file_name).execute()
                return data
            except errors.HttpError as error:
                if "Not Found" in error.content:
                    self.logger.info(
                        "File {} was not found in bucket {}".format(bucket_name, file_name))
                else:
                    raise error
        return {}

    def upload_file_to_bucket(self, bucket_name, file_path):
        def handle_progressless_iter(error, progressless_iters):
            if progressless_iters > NUM_RETRIES:
                self.logger.info('Failed to make progress for too many consecutive iterations.')
                raise error

            sleeptime = random.random() * (2 ** progressless_iters)
            self.logger.info(
                'Caught exception ({}). Sleeping for {} seconds before retry #{}.'.format(
                    str(error), sleeptime, progressless_iters))

            time.sleep(sleeptime)

        self.logger.info('Building upload request...')
        media = MediaFileUpload(file_path, chunksize=CHUNKSIZE, resumable=True)
        if not media.mimetype():
            media = MediaFileUpload(file_path, DEFAULT_MIMETYPE, resumable=True)

        blob_name = os.path.basename(file_path)
        if not self.bucket_exists(bucket_name):
            self.logger.error("Bucket {} doesn't exists".format(bucket_name))
            raise "Bucket doesn't exist"

        request = self._storage.objects().insert(
            bucket=bucket_name, name=blob_name, media_body=media)
        self.logger.info('Uploading file: {}, to bucket: {}, blob: {}'.format(
            file_path, bucket_name, blob_name))

        progressless_iters = 0
        response = None
        while response is None:
            error = None
            try:
                progress, response = request.next_chunk()
                if progress:
                    self.logger.info('Upload {}%'.format(100 * progress.progress()))
            except errors.HttpError as error:
                if error.resp.status < 500:
                    raise
            except RETRYABLE_ERRORS as error:
                if error:
                    progressless_iters += 1
                    handle_progressless_iter(error, progressless_iters)
                else:
                    progressless_iters = 0

        self.logger.info('Upload complete!')
        self.logger.info('Uploaded Object:')
        self.logger.info(json_dumps(response, indent=2))
        return (True, blob_name)

    def deploy_template(self, template, **kwargs):

        template_link = self.get_image_by_name(template)['selfLink']

        instance_name = kwargs['vm_name']
        self.logger.info("Creating {} instance".format(instance_name))

        machine_type = kwargs.get('machine_type',
            "zones/{}/machineTypes/n1-standard-1".format(self._zone))
        script = kwargs.get('startup_script_data', "#!/bin/bash")
        timeout = kwargs.get('timeout', 180)

        config = {
            'name': instance_name,
            'machineType': machine_type,

            # Specify the boot disk and the image to use as a source.
            'disks': [
                {
                    'boot': True,
                    'autoDelete': True,
                    'initializeParams': {
                        'sourceImage': template_link,
                    }
                }
            ],

            # Specify a network interface with NAT to access the public
            # internet.
            'networkInterfaces': [{
                'network': 'global/networks/default',
                'accessConfigs': [
                    {'type': 'ONE_TO_ONE_NAT', 'name': 'External NAT'}
                ]
            }],

            # Allow the instance to access cloud storage and logging.
            'serviceAccounts': [{
                'email': 'default',
                'scopes': [
                    'https://www.googleapis.com/auth/devstorage.read_write',
                    'https://www.googleapis.com/auth/logging.write'
                ]
            }],

            # Metadata is readable from the instance and allows you to
            # pass configuration from deployment scripts to instances.
            'metadata': {
                'items': [{
                    # Startup script is automatically executed by the
                    # instance upon startup.
                    'key': 'startup-script',
                    'value': script
                }, {
                    # Every project has a default Cloud Storage bucket that's
                    # the same name as the project.
                    'key': 'bucket',
                    'value': self._project
                }]
            },
            'tags': {
                'items': ['https-server']
            }
        }

        if kwargs.get('ssh_key', None):
            ssh_keys = {
                'key': 'ssh-keys',
                'value': kwargs.get('ssh_key', None)
            }
            config['metadata']['items'].append(ssh_keys)

        operation = self._instances.insert(
            project=self._project, zone=self._zone, body=config).execute()
        wait_for(lambda: self._nested_operation_wait(operation['name']), delay=0.5,
            num_sec=timeout, message=" Create {}".format(instance_name))
        return True

    def create_vm(self):
        raise NotImplementedError('create_vm not implemented.')

    def delete_vm(self, instance_name, timeout=180):
        if not self.does_vm_exist(instance_name):
            self.logger.info("The {} instance is not exists, skipping".format(instance_name))
            return True

        self.logger.info("Deleting Google Cloud instance {}".format(instance_name))
        operation = self._instances.delete(
            project=self._project, zone=self._zone, instance=instance_name).execute()
        wait_for(lambda: self._nested_operation_wait(operation['name']), delay=0.5,
            num_sec=timeout, message="Delete {}".format(instance_name))
        return True

    def restart_vm(self, instance_name):
        self.logger.info("Restarting Google Cloud instance {}".format(instance_name))
        operation = self._instances.reset(
            project=self._project, zone=self._zone, instance=instance_name).execute()
        wait_for(lambda: self._nested_operation_wait(operation['name']),
            message="Restart {}".format(instance_name))
        return True

    def stop_vm(self, instance_name):
        if self.is_vm_stopped(instance_name) or not self.does_vm_exist(instance_name):
            self.logger.info("The {} instance is already stopped or doesn't exist, skip termination"
               .format(instance_name))
            return True

        self.logger.info("Stoping Google Cloud instance {}".format(instance_name))
        operation = self._instances.stop(
            project=self._project, zone=self._zone, instance=instance_name).execute()
        wait_for(lambda: self._nested_operation_wait(operation['name']),
            message="Stop {}".format(instance_name), timeout=360)
        return True

    def start_vm(self, instance_name):
        # This method starts an instance that was stopped using the using the
        # instances().stop method.
        if self.is_vm_running(instance_name) or not self.does_vm_exist(instance_name):
            self.logger.info("The {} instance is already running or doesn't exists, skip starting"
               .format(instance_name))
            return True

        self.logger.info("Starting Google Cloud instance {}".format(instance_name))
        operation = self._instances.start(
            project=self._project, zone=self._zone, instance=instance_name).execute()
        wait_for(lambda: self._nested_operation_wait(operation['name']),
            message="Start {}".format(instance_name))
        return True

    def clone_vm(self, source_name, vm_name):
        raise NotImplementedError('clone_vm not implemented.')

    # Get external IP (ephemeral)
    def current_ip_address(self, vm_name):
        zones = self._compute.zones().list(project=self._project).execute()
        for zone in zones.get('items', []):
            zone_name = zone.get('name', None)
            for vm in self._get_zone_instances(zone_name).get('items', []):
                if vm['name'] == vm_name:
                    access_configs = vm.get('networkInterfaces')[0].get('accessConfigs')[0]
                    return access_configs.get('natIP')

    def disconnect(self):
        """Disconnect from the GCE

        GCE service is stateless, so there's nothing to disconnect from
        """
        pass

    def does_vm_exist(self, instance_name):
        try:
            self._find_instance_by_name(instance_name)
            return True
        except Exception:
            return False

    def does_forwarding_rule_exist(self, forwarding_rule_name):
        try:
            self._find_forwarding_rule_by_name(forwarding_rule_name)
            return True
        except Exception:
            return False

    def get_ip_address(self, vm_name):
        return self.current_ip_address(vm_name)

    def info(self):
        raise NotImplementedError('info not implemented.')

    def is_vm_running(self, vm_name):
        return self.vm_status(vm_name) in self.states['running']

    def is_vm_stopped(self, vm_name):
        return self.vm_status(vm_name) in self.states['stopped']

    def is_vm_suspended(self, vm_name):
        raise ActionNotSupported('vm_suspend not supported.')

    # These methods indicate if the vm is in the process of stopping or starting
    def is_vm_stopping(self, vm_name):
        return self.vm_status(vm_name) in self.states['stopping']

    def is_vm_starting(self, vm_name):
        return self.vm_status(vm_name) in self.states['starting']

    def list_flavor(self):
        raise NotImplementedError('list_flavor not implemented.')

    def list_template(self):
        return self.list_image()

    def remove_host_from_cluster(self, hostname):
        raise NotImplementedError('remove_host_from_cluster not implemented.')

    def suspend_vm(self, vm_name):
        raise ActionNotSupported('vm_suspend not supported.')

    def vm_status(self, vm_name):
        if self.does_vm_exist(vm_name):
            return self._find_instance_by_name(vm_name)['status']
        return None

    def wait_vm_running(self, vm_name, num_sec=360):
        self.logger.info("Waiting for instance {} to change status to ACTIVE".format(vm_name))
        wait_for(self.is_vm_running, [vm_name], num_sec=num_sec)

    def wait_vm_stopped(self, vm_name, num_sec=360):
        self.logger.info("Waiting for instance {} to change status to TERMINATED".format(vm_name))
        wait_for(self.is_vm_stopped, [vm_name], num_sec=num_sec)

    def wait_vm_suspended(self, vm_name, num_sec):
        raise ActionNotSupported('vm_suspend not supported.')

    def all_vms(self):
        result = []
        zones = self._compute.zones().list(project=self._project).execute()
        for zone in zones.get('items', []):
            zone_name = zone.get('name', None)
            for vm in self._get_zone_instances(zone_name).get('items', []):
                if vm['id'] and vm['name'] and vm['status'] and vm.get('networkInterfaces'):

                    result.append(VMInfo(
                        vm['id'],
                        vm['name'],
                        vm['status'],
                        vm.get('networkInterfaces')[0].get('networkIP'),
                    ))
        return result

    def vm_creation_time(self, instance_name):
        instance = self._find_instance_by_name(instance_name)
        vm_time_stamp = instance['creationTimestamp']
        creation_time = (iso8601.parse_date(vm_time_stamp)).replace(tzinfo=None)
        return creation_time

    def vm_type(self, instance_name):
        instance = self._find_instance_by_name(instance_name)
        if instance.get('machineType', None):
            return instance['machineType'].split('/')[-1]



# coding: utf-8
"""Backend management system classes

Used to communicate with providers without using CFME facilities
"""
from abc import ABCMeta, abstractmethod
from collections import namedtuple
from contextlib import contextmanager
from .exceptions import VMInstanceNotFound

from wait_for import wait_for, TimedOutError


VMInfo = namedtuple("VMInfo", ["uuid", "name", "power_state", "ip"])


class MgmtSystemAPIBase(object):
    """Base interface class for Management Systems

    Interface notes:

    * Initializers of subclasses must support \*\*kwargs in their
      signtures
    * Action methods (start/stop/etc) should block until the requested
      action is complete

    """
    __metaclass__ = ABCMeta
    STEADY_WAIT_MINS = 3

    # Flags to indicate whether or not this MgmtSystem can suspend/pause,
    can_suspend = True
    can_pause = False

    def __init__(self, kwargs):
        logger = kwargs.get('logger', None)
        self.logger = Logger(logger)

    @abstractmethod
    def start_vm(self, vm_name):
        """Starts a vm.

        Args:
            vm_name: name of the vm to be started
        Returns: whether vm action has been initiated properly
        """
        raise NotImplementedError('start_vm not implemented.')

    @abstractmethod
    def wait_vm_running(self, vm_name, num_sec):
        """Waits for a VM to be running.

        Args:
            vm_name: name of the vm to be running
            num_sec: number of seconds before timeout
        """
        raise NotImplementedError('wait_vm_running not implemented.')

    @abstractmethod
    def stop_vm(self, vm_name):
        """Stops a vm.

        Args:
            vm_name: name of the vm to be stopped
        Returns: whether vm action has been initiated properly
        """
        raise NotImplementedError('stop_vm not implemented.')

    @abstractmethod
    def wait_vm_stopped(self, vm_name, num_sec):
        """Waits for a VM to be stopped.

        Args:
            vm_name: name of the vm to be stopped
            num_sec: number of seconds before timeout
        """
        raise NotImplementedError('wait_vm_stopped not implemented.')

    @abstractmethod
    def create_vm(self, vm_name, *args, **kwargs):
        """Creates a vm.

        Args:
            vm_name: name of the vm to be created
        Returns: whether vm action has been initiated properly
        """
        raise NotImplementedError('create_vm not implemented.')

    @abstractmethod
    def delete_vm(self, vm_name, *args, **kwargs):
        """Deletes a vm.

        Args:
            vm_name: name of the vm to be deleted
        Returns: whether vm action has been initiated properly
        """
        raise NotImplementedError('delete_vm not implemented.')

    @abstractmethod
    def restart_vm(self, vm_name):
        """Restart a vm.

        Args:
            vm_name: name of the vm to be restarted
        Returns: whether vm stop/start have been initiated properly
        """
        raise NotImplementedError('restart_vm not implemented.')

    @abstractmethod
    def list_vm(self, **kwargs):
        """Returns a list of vm names.

        Returns: list of vm names
        """
        raise NotImplementedError('list_vm not implemented.')

    @abstractmethod
    def list_template(self):
        """Returns a list of templates/images.

        Returns: list of template/image names
        """
        raise NotImplementedError('list_template not implemented.')

    @abstractmethod
    def list_flavor(self):
        """Returns a list of flavors.

        Only valid for OpenStack and Amazon

        Returns: list of flavor names
        """
        raise NotImplementedError('list_flavor not implemented.')

    def list_network(self):
        """Returns a list of networks.

        Only valid for OpenStack

        Returns: list of network names
        """
        raise NotImplementedError('list_network not implemented.')

    @abstractmethod
    def info(self):
        """Returns basic information about the mgmt system.

        Returns: string representation of name/version of mgmt system.
        """
        raise NotImplementedError('info not implemented.')

    @abstractmethod
    def disconnect(self):
        """Disconnects the API from mgmt system"""
        raise NotImplementedError('disconnect not implemented.')

    @abstractmethod
    def vm_status(self, vm_name):
        """Status of VM.

        Args:
            vm_name: name of the vm to get status
        Returns: state of the vm
        """
        raise NotImplementedError('vm_status not implemented.')

    @abstractmethod
    def is_vm_running(self, vm_name):
        """Is the vm running?

        Args:
            vm_name: name of the vm
        Returns: whether the vm is running or not
        """
        raise NotImplementedError('is_vm_running not implemented.')

    @abstractmethod
    def is_vm_stopped(self, vm_name):
        """Is the vm stopped?

        Args:
            vm_name: name of the vm
        Returns: whether the vm is stopped or not
        """
        raise NotImplementedError('is_vm_stopped not implemented.')

    @abstractmethod
    def is_vm_suspended(self, vm_name):
        """Is the vm suspended?

        Args:
            vm_name: name of the vm
        Returns: whether the vm is suspended or not
        """
        raise NotImplementedError('is_vm_suspended not implemented.')

    @abstractmethod
    def suspend_vm(self, vm_name):
        """Suspend a vm.

        Args:
            vm_name: name of the vm to be suspended
        Returns: whether vm suspend has been initiated properly
        """
        raise NotImplementedError('restart_vm not implemented.')

    @abstractmethod
    def wait_vm_suspended(self, vm_name, num_sec):
        """Waits for a VM to be suspended.

        Args:
            vm_name: name of the vm to be suspended
            num_sec: number of seconds before timeout
        """
        raise NotImplementedError('wait_vm_suspended not implemented.')

    @abstractmethod
    def clone_vm(self, source_name, vm_name):
        """Clone a VM.

        Args:
            source_name: The source VM to clone from
            vm_name: The name of the new VM
        Returns: IP address of the clone
        """
        raise NotImplementedError('clone_vm not implemented.')

    @abstractmethod
    def does_vm_exist(self, name):
        """Does VM exist?

        Args:
            vm_name: The name of the VM
        Returns: whether vm exists
        """
        raise NotImplementedError('does_vm_exist not implemented.')

    @abstractmethod
    def deploy_template(self, template, *args, **kwargs):
        """Deploy a VM from a template

        Args:
            template: The name of the template to deploy
        Returns: name or id(ec2) of vm
        """
        raise NotImplementedError('deploy_template not implemented.')

    @abstractmethod
    def current_ip_address(self, vm_name):
        """Returns current IP address. Returns None if the address could not have been determined.

        Args:
            vm_name: The name of the VM
        Returns: vm ip address or None
        """
        raise NotImplementedError('current_ip_address not implemented.')

    @abstractmethod
    def get_ip_address(self, vm_name):
        """get VM ip address - blocks until the waiting is finished

        Args:
            vm_name: The name of the VM
        Returns: vm ip address
        """
        raise NotImplementedError('get_ip_address not implemented.')

    @abstractmethod
    def remove_host_from_cluster(self, hostname):
        """remove a host from it's cluster

        :param hostname: The hostname of the system
        :type  hostname: str
        :return: True if successful, False if failed
        :rtype: boolean

        """
        raise NotImplementedError('remove_host_from_cluster not implemented.')

    def stats(self, *requested_stats):
        """Returns all available stats, if none are explicitly requested

        Args:
            *requested_stats: A list giving the name of the stats to return. Stats are defined
                in the _stats_available attibute of the specific class.
        Returns: A dict of stats.
        """
        requested_stats = requested_stats or self._stats_available
        return {stat: self._stats_available[stat](self) for stat in requested_stats}

    def in_steady_state(self, vm_name):
        """Return whether the specified virtual machine is in steady state

        Args:
            vm_name: VM name
        Returns: boolean
        """
        return (self.is_vm_running(vm_name) or self.is_vm_stopped(vm_name) or
            self.is_vm_suspended(vm_name))

    def wait_vm_steady(self, vm_name):
        """Waits 2 (or user-specified time) minutes for VM to settle in steady state

        Args:
            vm_name: VM name
        """
        try:
            return wait_for(
                lambda: self.in_steady_state(vm_name),
                num_sec=self.STEADY_WAIT_MINS * 60,
                delay=2,
                message="VM %s in steady state" % vm_name
            )
        except TimedOutError:
            self.logger.exception(
                "VM {} got stuck in {} state when waiting for steady state.".format(
                    vm_name, self.vm_status(vm_name)))
            raise

    @property
    def can_rename(self):
        return hasattr(self, "rename_vm")

    @contextmanager
    def steady_wait(self, minutes):
        """Overrides original STEADY_WAIT_MINS variable in the object.

        This is useful eg. when creating templates in RHEV as it has long Image Locked period

        Args:
            minutes: How many minutes to wait
        """
        original = None
        if "STEADY_WAIT_MINS" in self.__dict__:
            original = self.__dict__["STEADY_WAIT_MINS"]
        self.__dict__["STEADY_WAIT_MINS"] = minutes
        yield
        if original is None:
            del self.__dict__["STEADY_WAIT_MINS"]
        else:
            self.__dict__["STEADY_WAIT_MINS"] = original

    def does_template_exist(self, template_name):
        """If system does not implement anything better, this will work """
        return template_name in self.list_template()

    def delete_template(self, template_name):
        return self.delete_vm(template_name)  # Fall back to original vSphere behaviour

    def set_meta_value(self, instance, key, value):
        raise NotImplementedError(
            'Provider {} does not implement set_meta_value'.format(type(self).__name__))

    def get_meta_value(self, instance, key):
        raise NotImplementedError(
            'Provider {} does not implement get_meta_value'.format(type(self).__name__))

    def get_vm_guid(self, vm_name):
        for vm in self.all_vms():
            if vm.name == vm_name:
                return vm.uuid
        else:
            raise VMInstanceNotFound(vm_name)


class ContainerMgmtSystemAPIBase(MgmtSystemAPIBase):
    """Base interface class for Container Management Systems

    Interface notes:

    * Initializers of subclasses must support \*\*kwargs in their
      signtures
    * Action methods (start/stop/etc) should block until the requested
      action is complete

    """

    def clone_vm(self, source_name, vm_name):
        raise NotImplementedError('clone_vm not implemented.')

    def create_vm(self, vm_name):
        raise NotImplementedError('create_vm not implemented.')

    def current_ip_address(self, vm_name):
        raise NotImplementedError('current_ip_address not implemented.')

    def delete_vm(self, vm_name):
        raise NotImplementedError('delete_vm not implemented.')

    def deploy_template(self, template, *args, **kwargs):
        raise NotImplementedError('deploy_template not implemented.')

    def disconnect(self):
        raise NotImplementedError('disconnect not implemented.')

    def does_vm_exist(self, name):
        raise NotImplementedError('does_vm_exist not implemented.')

    def get_ip_address(self, vm_name):
        raise NotImplementedError('get_ip_address not implemented.')

    def is_vm_running(self, vm_name):
        raise NotImplementedError('is_vm_running not implemented.')

    def is_vm_stopped(self, vm_name):
        raise NotImplementedError('is_vm_stopped not implemented.')

    def is_vm_suspended(self, vm_name):
        raise NotImplementedError('is_vm_suspended not implemented.')

    def list_flavor(self):
        raise NotImplementedError('list_flavor not implemented.')

    def list_template(self):
        raise NotImplementedError('list_template not implemented.')

    def list_vm(self, **kwargs):
        raise NotImplementedError('list_vm not implemented.')

    def remove_host_from_cluster(self, hostname):
        raise NotImplementedError('remove_host_from_cluster not implemented.')

    def restart_vm(self, vm_name):
        raise NotImplementedError('restart_vm not implemented.')

    def start_vm(self, vm_name):
        raise NotImplementedError('start_vm not implemented.')

    def stop_vm(self, vm_name):
        raise NotImplementedError('stop_vm not implemented.')

    def suspend_vm(self, vm_name):
        raise NotImplementedError('restart_vm not implemented.')

    def vm_status(self, vm_name):
        raise NotImplementedError('vm_status not implemented.')

    def wait_vm_running(self, vm_name, num_sec):
        raise NotImplementedError('wait_vm_running not implemented.')

    def wait_vm_stopped(self, vm_name, num_sec):
        raise NotImplementedError('wait_vm_stopped not implemented.')

    def wait_vm_suspended(self, vm_name, num_sec):
        raise NotImplementedError('wait_vm_suspended not implemented.')


class Logger(object):
    """ A simple logger proxy object

    If there is no logger, then we just return a catch_all function for the
    logger that accepts all args and does nothing, else we try to call the
    attribute on the real logger object that is passed in.
    """

    def __init__(self, logger=None):
        self.logger = logger

    def _catch_all(self, *args, **kwargs):
        return None

    def __getattr__(self, name):
        if self.logger:
            return getattr(self.logger, name)
        else:
            return self._catch_all



# coding: utf-8
"""Backend management system classes

Used to communicate with providers without using CFME facilities
"""
import fauxfactory
from ovirtsdk.api import API
from ovirtsdk.infrastructure.errors import DisconnectedError, RequestError
from ovirtsdk.xml import params
from wait_for import wait_for, TimedOutError

from base import MgmtSystemAPIBase, VMInfo
from exceptions import VMInstanceNotFound, VMInstanceNotSuspended, VMNotFoundViaIP


class RHEVMSystem(MgmtSystemAPIBase):
    """
    Client to RHEVM API

    This class piggy backs off ovirtsdk.

    Benefits of ovirtsdk:

    * Don't need intimite knowledge w/ RHEVM api itself.

    Detriments of ovirtsdk:

    * Response to most quaries are returned as an object rather than a string.
      This makes it harder to do simple stuff like getting the status of a vm.
    * Because of this, it makes listing VMs based on \*\*kwargs impossible
      since ovirtsdk relies on re class to find matches.

      * | For example: List out VM with this name (positive case)
        | Ideal: self.api.vms.list(name='test_vm')
        | Underneath the hood:

        * ovirtsdk fetches list of all vms [ovirtsdk.infrastructure.brokers.VM
          object, ...]
        * ovirtsdk then tries to filter the result using re.

          * tries to look for 'name' attr in ovirtsdk.infrastructure.brokers.VM
            object
          * found name attribute, in this case, the type of the value of the
            attribute is string.
          * match() succeed in comparing the value to 'test_vm'

      * | For example: List out VM with that's powered on (negative case)
        | Ideal: self.api.vms.list(status='up')
        | Underneath the hood:

        * **same step as above except**

          * found status attribute, in this case, the type of the value of
            the attribute is ovirtsdk.xml.params.Status
          * match() failed because class is compared to string 'up'

     This problem should be attributed to how RHEVM api was designed rather
     than how ovirtsdk handles RHEVM api responses.

    * Obj. are not updated after action calls.

      * For example::
          vm = api.vms.get(name='test_vm')
          vm.status.get_state() # returns 'down'
          vm.start()
          # wait a few mins
          vm.status.get_state() # returns 'down'; wtf?

          vm = api.vms.get(name='test_vm')
          vm.status.get_state() # returns 'up'

    Args:
        hostname: The hostname of the system.
        username: The username to connect with.
        password: The password to connect with.

    Keywords:
        port: (Optional) Port where RHEVM API listens.
        api_endpoint: (Optional) If you need to fine-tune and pass an exact endpoint in form of a
            full URL, use this keyword. the ``port`` keyword is then not used.

    Returns: A :py:class:`RHEVMSystem` object.
    """

    _stats_available = {
        'num_vm': lambda self: self.api.get_summary().get_vms().total,
        'num_host': lambda self: len(self.list_host()),
        'num_cluster': lambda self: len(self.list_cluster()),
        'num_template': lambda self: len(self.list_template()),
        'num_datastore': lambda self: len(self.list_datastore()),
    }

    STEADY_WAIT_MINS = 6

    def __init__(self, hostname, username, password, **kwargs):
        # generate URL from hostname
        super(RHEVMSystem, self).__init__(kwargs)
        url_component = 'ovirt-engine/api' if float(kwargs['version']) >= 4.0 else 'api'
        if 'api_endpoint' in kwargs:
            url = kwargs['api_endpoint']
        elif 'port' in kwargs:
            url = 'https://{}:{}/{}'.format(hostname, kwargs['port'], url_component)
        else:
            url = 'https://{}/{}'.format(hostname, url_component)

        self._api = None
        self._api_kwargs = {
            'url': url,
            'username': username,
            'password': password,
            'insecure': True
        }
        self.kwargs = kwargs

    @property
    def api(self):
        # test() will return false if the connection timeouts, catch it and force it to re-init
        try:
            if self._api is None or (self._api is not None and not self._api.test()):
                self._api = API(**self._api_kwargs)
        # if the connection was disconnected, force it to re-init
        except DisconnectedError:
            self._api = API(**self._api_kwargs)
        return self._api

    def _get_vm(self, vm_name=None):
        """ Returns a vm from the RHEVM object.

        Args:
            vm_name: The name of the VM.

        Returns: an ovirtsdk vm object.
        """
        if vm_name is None:
            raise VMInstanceNotFound(vm_name)
        else:
            vm = self.api.vms.get(name=vm_name)
            if vm is None:
                raise VMInstanceNotFound(vm_name)
            return vm

    def current_ip_address(self, vm_name):
        info = self._get_vm(vm_name).get_guest_info()
        if info is None:
            return None
        try:
            return info.get_ips().get_ip()[0].get_address()
        except (AttributeError, IndexError):
            return None

    def get_ip_address(self, vm_name, timeout=600):
        try:
            return wait_for(lambda: self.current_ip_address(vm_name),
                fail_condition=None, delay=5, num_sec=timeout,
                message="get_ip_address from rhevm")[0]
        except TimedOutError:
            return None

    def get_vm_name_from_ip(self, ip):
        # unfortunately it appears you cannot query for ip address from the sdk,
        #   unlike curling rest api which does work
        """ Gets the name of a vm from its IP.

        Args:
            ip: The ip address of the vm.
        Returns: The vm name for the corresponding IP."""

        vms = self.api.vms.list()

        for vm in vms:
            if vm.get_guest_info() is None or vm.get_guest_info().get_ips() is None:
                continue
            else:
                for addr in vm.get_guest_info().get_ips().get_ip():
                    if ip in addr.get_address():
                        return vm.name
        raise VMNotFoundViaIP('The requested IP is not known as a VM')

    def does_vm_exist(self, name):
        try:
            self._get_vm(name)
            return True
        except VMInstanceNotFound:
            return False

    def start_vm(self, vm_name=None):
        self.wait_vm_steady(vm_name)
        self.logger.info(' Starting RHEV VM %s' % vm_name)
        vm = self._get_vm(vm_name)
        if vm.status.get_state() == 'up':
            self.logger.info(' RHEV VM %s os already running.' % vm_name)
            return True
        else:
            vm.start()
            self.wait_vm_running(vm_name)
            return True

    def stop_vm(self, vm_name):
        self.wait_vm_steady(vm_name)
        self.logger.info(' Stopping RHEV VM %s' % vm_name)
        if self.is_vm_stopped(vm_name):
            self.logger.info(' RHEV VM %s os already stopped.' % vm_name)
            return True
        else:
            vm = self._get_vm(vm_name)
            vm.stop()
            self.wait_vm_stopped(vm_name)
            return True

    def delete_vm(self, vm_name):
        self.wait_vm_steady(vm_name)
        if not self.is_vm_stopped(vm_name):
            self.stop_vm(vm_name)
        self.logger.debug(' Deleting RHEV VM %s' % vm_name)

        def _do_delete():
            """Returns True if you have to retry"""
            if not self.does_vm_exist(vm_name):
                return False
            try:
                vm = self._get_vm(vm_name)
                vm.delete()
            except RequestError as e:
                # Handle some states that can occur and can be circumvented
                if e.status == 409 and "Related operation" in e.detail:
                    self.logger.info("Waiting for RHEV: {}:{} ({})".format(
                        e.status, e.reason, e.detail))
                    return True
                else:
                    raise  # Raise other so we can see them and eventually add them into handling
                # TODO: handle 400 - but I haven't seen the error message, it was empty.
            else:
                return False

        wait_for(_do_delete, fail_condition=True, num_sec=600, delay=15, message="execute delete")

        wait_for(
            lambda: self.does_vm_exist(vm_name),
            fail_condition=True,
            message="wait for RHEV VM %s deleted" % vm_name,
            num_sec=300
        )
        return True

    def create_vm(self, vm_name):
        raise NotImplementedError('This function has not yet been implemented.')
    # Heres the code but don't have a need and no time to test it to get it right
    #   including for inclusion later
    #
    # def create_vm(self, vm_name, *args, **kwargs):
    #     MB = 1024 * 1024
    #     try:
    #         self.api.vms.add(
    #             params.VM(
    #                 name=vm_name,
    #                 memory=kwargs['memory_in_mb'] * MB,
    #                 cluster=self.api.clusters.get(kwargs['cluster_name']),
    #                 template=self.api.templates.get('Blank')))
    #         print 'VM created'
    #         self.api.vms.get(vm_name).nics.add(params.NIC(name='eth0',
    #             network=params.Network(name='ovirtmgmt'), interface='virtio'))
    #         print 'NIC added to VM'
    #         self.api.vms.get(vm_name).disks.add(params.Disk(
    #             storage_domains=params.StorageDomains(
    #                 storage_domain=[self.api.storagedomains.get(kwargs['storage_domain'])],
    #                 size=512 * MB,
    #                 status=None,
    #                 interface='virtio',
    #                 format='cow',
    #                 sparse=True,
    #                 bootable=True)))
    #         print 'Disk added to VM'
    #         print 'Waiting for VM to reach Down status'
    #         while self.api.vms.get(vm_name).status.state != 'down':
    #             time.sleep(1)
    #     except Exception as e:
    #         print 'Failed to create VM with disk and NIC\n%s' % str(e)

    def restart_vm(self, vm_name):
        self.logger.debug(' Restarting RHEV VM %s' % vm_name)
        return self.stop_vm(vm_name) and self.start_vm(vm_name)

    def list_vm(self, **kwargs):
        # list vm based on kwargs can be buggy
        # i.e. you can't return a list of powered on vm
        # but you can return a vm w/ a matched name
        vm_list = self.api.vms.list(**kwargs)
        return [vm.name for vm in vm_list]

    def all_vms(self):
        result = []
        for vm in self.api.vms.list():
            try:
                ip = vm.get_guest_info().get_ips().get_ip()[0].get_address()
            except (AttributeError, IndexError):
                ip = None
            result.append(
                VMInfo(
                    vm.get_id(),
                    vm.get_name(),
                    vm.get_status().get_state(),
                    ip,
                )
            )
        return result

    def get_vm_guid(self, vm_name):
        return self._get_vm(vm_name).get_id()

    def list_host(self, **kwargs):
        host_list = self.api.hosts.list(**kwargs)
        return [host.name for host in host_list]

    def list_datastore(self, **kwargs):
        datastore_list = self.api.storagedomains.list(**kwargs)
        return [ds.name for ds in datastore_list if ds.get_status() is None]

    def list_cluster(self, **kwargs):
        cluster_list = self.api.clusters.list(**kwargs)
        return [cluster.name for cluster in cluster_list]

    def list_template(self, **kwargs):
        """
        Note: CFME ignores the 'Blank' template, so we do too
        """
        template_list = self.api.templates.list(**kwargs)
        return [template.name for template in template_list if template.name != "Blank"]

    def list_flavor(self):
        raise NotImplementedError('This function is not supported on this platform.')

    def info(self):
        # and we got nothing!
        pass

    def disconnect(self):
        self.api.disconnect()

    def vm_status(self, vm_name=None):
        return self._get_vm(vm_name).get_status().get_state()

    def vm_creation_time(self, vm_name):
        vm = self._get_vm(vm_name)
        return vm.get_creation_time().replace(tzinfo=None)

    def in_steady_state(self, vm_name):
        return self.vm_status(vm_name) in {"up", "down", "suspended"}

    def is_vm_running(self, vm_name):
        return self.vm_status(vm_name) == "up"

    def wait_vm_running(self, vm_name, num_sec=360):
        self.logger.info(" Waiting for RHEV-M VM %s to change status to ON" % vm_name)
        wait_for(self.is_vm_running, [vm_name], num_sec=num_sec)

    def is_vm_stopped(self, vm_name):
        return self.vm_status(vm_name) == "down"

    def wait_vm_stopped(self, vm_name, num_sec=360):
        self.logger.info(" Waiting for RHEV-M VM %s to change status to OFF" % vm_name)
        wait_for(self.is_vm_stopped, [vm_name], num_sec=num_sec)

    def is_vm_suspended(self, vm_name):
        return self.vm_status(vm_name) == "suspended"

    def wait_vm_suspended(self, vm_name, num_sec=720):
        self.logger.info(" Waiting for RHEV-M VM %s to change status to SUSPENDED" % vm_name)
        wait_for(self.is_vm_suspended, [vm_name], num_sec=num_sec)

    def suspend_vm(self, vm_name):
        self.wait_vm_steady(vm_name)
        self.logger.debug(' Suspending RHEV VM %s' % vm_name)
        vm = self._get_vm(vm_name)
        if vm.status.get_state() == 'down':
            raise VMInstanceNotSuspended(vm_name)
        elif vm.status.get_state() == 'suspended':
            self.logger.info(' RHEV VM %s is already suspended.' % vm_name)
            return True
        else:
            vm.suspend()
            self.wait_vm_suspended(vm_name)
            return True

    def clone_vm(self, source_name, vm_name):
        raise NotImplementedError('This function has not yet been implemented.')

    def deploy_template(self, template, *args, **kwargs):
        self.logger.debug(' Deploying RHEV template %s to VM %s' % (template, kwargs["vm_name"]))
        timeout = kwargs.pop('timeout', 900)
        power_on = kwargs.pop('power_on', True)
        vm_kwargs = {
            'name': kwargs['vm_name'],
            'cluster': self.api.clusters.get(kwargs['cluster']),
            'template': self.api.templates.get(template)
        }
        if 'placement_policy_host' in kwargs and 'placement_policy_affinity' in kwargs:
            host = params.Host(name=kwargs['placement_policy_host'])
            policy = params.VmPlacementPolicy(host=host,
                affinity=kwargs['placement_policy_affinity'])
            vm_kwargs['placement_policy'] = policy
        if 'cpu' in kwargs:
            vm_kwargs['cpu'] = params.CPU(topology=params.CpuTopology(cores=int(kwargs['cpu'])))
        if 'ram' in kwargs:
            vm_kwargs['memory'] = int(kwargs['ram']) * 1024 * 1024  # MB
        self.api.vms.add(params.VM(**vm_kwargs))
        self.wait_vm_stopped(kwargs['vm_name'], num_sec=timeout)
        if power_on:
            self.start_vm(kwargs['vm_name'])
        return kwargs['vm_name']

    def remove_host_from_cluster(self, hostname):
        raise NotImplementedError('remove_host_from_cluster not implemented')

    def mark_as_template(
            self, vm_name, delete=True, temporary_name=None, delete_on_error=True, **kwargs):
        """Turns the VM off, creates template from it and deletes the original VM.

        Mimics VMware behaviour here.

        Args:
            vm_name: Name of the VM to be turned to template
            delete: Whether to delete the VM (default: True)
            temporary_name: If you want, you can specific an exact temporary name for renaming.
        """
        temp_template_name = temporary_name or "templatize_{}".format(
            fauxfactory.gen_alphanumeric(8))
        try:
            with self.steady_wait(30):
                create_new_template = True
                if self.does_template_exist(temp_template_name):
                    try:
                        self._wait_template_ok(temp_template_name)
                    except VMInstanceNotFound:
                        pass  # It got deleted.
                    else:
                        create_new_template = False
                        if self.does_vm_exist(vm_name) and delete:
                            self.delete_vm(vm_name)
                        if delete:  # We can only rename to the original name if we delete the vm
                            self._rename_template(temp_template_name, vm_name)

                if create_new_template:
                    self.stop_vm(vm_name)
                    vm = self._get_vm(vm_name)
                    actual_cluster = vm.get_cluster()
                    new_template = params.Template(
                        name=temp_template_name, vm=vm, cluster=actual_cluster)
                    self.api.templates.add(new_template)
                    # First it has to appear
                    self._wait_template_exists(temp_template_name)
                    # Then the process has to finish
                    self._wait_template_ok(temp_template_name)
                    # Delete the original VM
                    if self.does_vm_exist(vm_name) and delete:
                        self.delete_vm(vm_name)
                    if delete:  # We can only rename to the original name if we delete the vm
                        self._rename_template(temp_template_name, vm_name)
        except TimedOutError:
            if delete_on_error:
                self.delete_template(temp_template_name)
            raise

    def _rename_template(self, old_name, new_name):
        template = self.api.templates.get(name=old_name)
        if template is None:
            raise VMInstanceNotFound("Template {} not found!".format(old_name))
        template.set_name(new_name)
        template.update()

    def rename_vm(self, vm_name, new_vm_name):
        vm = self._get_vm(vm_name)
        try:
            vm.set_name(new_vm_name)
            vm.update()
        except Exception as e:
            self.logger.exception(e)
            return vm_name
        else:
            return new_vm_name

    def _wait_template_ok(self, template_name):
        try:
            wait_for(
                lambda:
                self.api.templates.get(name=template_name).get_status().state == "ok",
                num_sec=30 * 60, message="template is OK", delay=45)
        except AttributeError:  # .get() returns None when template not found
            raise VMInstanceNotFound("Template {} not found!".format(template_name))

    def _wait_template_exists(self, template_name):
        wait_for(
            lambda: self.does_template_exist(template_name),
            num_sec=30 * 60, message="template exists", delay=45)

    def does_template_exist(self, template_name):
        return self.api.templates.get(name=template_name) is not None

    def delete_template(self, template_name):
        template = self.api.templates.get(name=template_name)
        if template is None:
            self.logger.info(
                " Template {} is already not present on the RHEV-M provider".format(template_name))
            return
        self._wait_template_ok(template_name)
        template.delete()
        wait_for(
            lambda: not self.does_template_exist(template_name),
            num_sec=15 * 60, delay=20)



import requests
import os
import json
from exceptions import RestClientException

requests.packages.urllib3.disable_warnings()


class BearerTokenAuth(requests.auth.AuthBase):
    """Attaches a bearer token to the given request object"""
    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers['Authorization'] = 'Bearer {}'.format(self.token)
        return r


class ContainerClient(object):

    def __init__(self, hostname, auth, protocol="https", port=6443, entry='api/v1', verify=False):
        """Simple REST API client for container management systems

        Args:
            hostname: String with the hostname or IP address of the server (e.g. '10.11.12.13')
            auth: Either a (user, pass) sequence or a string with token
            protocol: Protocol to use for communication with the server
            port: Port to use
            entry: Entry point of the REST API
            verify: 'True' if we want to verify SSL, 'False' otherwise
        """
        self.api_entry = "{}://{}:{}/{}".format(protocol, hostname, port, entry)
        self.verify = verify
        if type(auth) in (list, set, tuple):
            self.auth = auth
        elif type(auth) is str:
            self.auth = BearerTokenAuth(auth)
        else:
            raise RestClientException('Invalid auth object')

    def get(self, entity_type, name=None, namespace=None):
        """Sends a request to fetch an entity of specific type

        Fetches a single entity if its name is provided or all of given type if name is ommited.

        Note:
            Some entities are tied to namespaces (projects).
            To fetch these by name, namespace has to be provided as well.

        Return:
            Tuple containing status code and json response with requested entity/entities.
        """
        path = '{}s'.format(entity_type)
        if name is not None:
            if namespace is not None:
                path = os.path.join('namespaces/{}'.format(namespace), path)
            path = os.path.join(path, '{}'.format(name))
        r = self.raw_get(path)
        return (r.status_code, r.json() if r.ok else None)

    def get_json(self, path, headers=None, params=None):
        r = self.raw_get(path, headers, params)
        return (r.json() if r.ok else None)

    def put_status(self, path, data, headers=None):
        r = self.raw_put(path, data, headers)
        return r.ok

    def post_status(self, path, data, headers=None):
        r = self.raw_post(path, data, headers)
        return r.ok

    def delete_status(self, path, headers=None):
        r = self.raw_delete(path, headers)
        return r.ok

    def raw_get(self, path, headers=None, params=None):
        return requests.get(
            os.path.join(self.api_entry, path),
            auth=self.auth,
            verify=self.verify,
            headers=headers,
            params=params)

    def raw_put(self, path, data, headers=None):
        return requests.put(
            os.path.join(self.api_entry, path), auth=self.auth, verify=self.verify,
            headers=headers, data=json.dumps(data))

    def raw_post(self, path, data, headers=None):
        return requests.post(
            os.path.join(self.api_entry, path), auth=self.auth, verify=self.verify,
            headers=headers, data=json.dumps(data))

    def raw_delete(self, path, headers=None):
        return requests.delete(
            os.path.join(self.api_entry, path), auth=self.auth, verify=self.verify,
            headers=headers)



from collections import namedtuple
from kubernetes import Kubernetes, ImageRegistry, Project, Service
from rest_client import ContainerClient

"""
Related yaml structures:

[cfme_data]
management_systems:
    openshift:
        name: My openshift
        type: openshift
        hostname: 10.12.13.14
        port: 8443
        credentials: openshift
        authenticate: true
        rest_protocol: https

[credentials]
openshift:
    username: admin
    password: secret
    token: mytoken
"""

Route = namedtuple('Route', ['name', 'project_name'])


class Openshift(Kubernetes):

    _stats_available = Kubernetes._stats_available.copy()
    _stats_available.update({
        'num_route': lambda self: len(self.list_route())
    })

    def __init__(self,
            hostname, protocol="https", port=8443, k_entry="api/v1", o_entry="oapi/v1", **kwargs):
        self.hostname = hostname
        self.username = kwargs.get('username', '')
        self.password = kwargs.get('password', '')
        self.token = kwargs.get('token', '')
        self.auth = self.token if self.token else (self.username, self.password)
        self.k_api = ContainerClient(hostname, self.auth, protocol, port, k_entry)
        self.o_api = ContainerClient(hostname, self.auth, protocol, port, o_entry)
        self.api = self.k_api  # default api is the kubernetes one for Kubernetes-class requests

    def list_route(self):
        """Returns list of routes"""
        entities = []
        entities_j = self.o_api.get('route')[1]['items']
        for entity_j in entities_j:
            meta = entity_j['metadata']
            entity = Route(meta['name'], meta['namespace'])
            entities.append(entity)
        return entities

    def list_service(self):
        """Returns list of services"""
        entities = []
        entities_j = self.api.get('service')[1]['items']
        for entity_j in entities_j:
            meta, spec = entity_j['metadata'], entity_j['spec']
            entity = Service(
                meta['name'], meta['namespace'], spec['clusterIP'], spec['sessionAffinity'])
            entities.append(entity)
        return entities

    def list_image_registry(self):
        """Returns list of image registries (image streams)"""
        entities = []
        entities_j = self.o_api.get('imagestream')[1]['items']
        for entity_j in entities_j:
            if 'dockerImageRepository' not in entity_j['status']:
                continue
            reg_raw = entity_j['status']['dockerImageRepository'].split('/')[0]
            host, port = reg_raw.split(':') if ':' in reg_raw else (reg_raw, '')
            entity = ImageRegistry(host, port)
            if entity not in entities:
                entities.append(entity)
        return entities

    def list_project(self):
        """Returns list of projects"""
        entities = []
        entities_j = self.o_api.get('project')[1]['items']
        for entity_j in entities_j:
            meta = entity_j['metadata']
            entity = Project(meta['name'])
            entities.append(entity)
        return entities



# coding: utf-8
from keystoneclient.v2_0 import client as oskclient
from novaclient import client as osclient
from novaclient.client import HTTPClient
from requests.exceptions import Timeout

from base import MgmtSystemAPIBase


# TODO The following monkeypatch nonsense is criminal, and would be
# greatly simplified if openstack made it easier to specify a custom
# client class. This is a trivial PR that they're likely to accept.

# Note: This same mechanism may be required for keystone and cinder
# clients, but hopefully won't be.

# monkeypatch method to add retry support to openstack
def _request_timeout_handler(self, url, method, retry_count=0, **kwargs):
    try:
        # Use the original request method to do the actual work
        return HTTPClient.request(self, url, method, **kwargs)
    except Timeout:
        if retry_count >= 3:
            self._cfme_logger.error('nova request timed out after {} retries'.format(retry_count))
            raise
        else:
            # feed back into the replaced method that supports retry_count
            retry_count += 1
            self._cfme_logger.error('nova request timed out; retry {}'.format(retry_count))
            return self.request(url, method, retry_count=retry_count, **kwargs)


class OpenstackInfraSystem(MgmtSystemAPIBase):
    """Openstack Infrastructure management system

    # TODO
    """

    _stats_available = {
        'num_template': lambda self: len(self.list_template()),
        'num_host': lambda self: len(self.list_host()),
    }

    states = {
        'running': ('ACTIVE',),
        'stopped': ('SHUTOFF',),
        'suspended': ('SUSPENDED',),
    }

    can_suspend = True

    def __init__(self, **kwargs):
        super(OpenstackInfraSystem, self).__init__(kwargs)
        self.tenant = kwargs['tenant']
        self.username = kwargs['username']
        self.password = kwargs['password']
        self.auth_url = kwargs['auth_url']
        self._api = None
        self._kapi = None
        self._capi = None

    @property
    def api(self):
        if not self._api:
            self._api = osclient.Client('2',
                                        self.username,
                                        self.password,
                                        self.tenant,
                                        self.auth_url,
                                        service_type="compute",
                                        insecure=True,
                                        timeout=30)
            # replace the client request method with our version that
            # can handle timeouts; uses explicit binding (versus
            # replacing the method directly on the HTTPClient class)
            # so we can still call out to HTTPClient's original request
            # method in the timeout handler method
            self._api.client._cfme_logger = self.logger
            self._api.client.request = _request_timeout_handler.__get__(self._api.client,
                HTTPClient)
        return self._api

    @property
    def kapi(self):
        if not self._kapi:
            self._kapi = oskclient.Client(username=self.username,
                                          password=self.password,
                                          tenant_name=self.tenant,
                                          auth_url=self.auth_url,
                                          insecure=True)
        return self._kapi

    @property
    def nodes(self):
        return self.api.servers.list()

    @property
    def images(self):
        return self.api.images.list()

    @property
    def networks(self):
        return self.api.networks.list()

    def start_vm(self, vm_name):
        raise NotImplementedError('start_vm not implemented.')

    def wait_vm_running(self, vm_name, num_sec):
        raise NotImplementedError('wait_vm_running not implemented.')

    def stop_vm(self, vm_name):
        raise NotImplementedError('stop_vm not implemented.')

    def wait_vm_stopped(self, vm_name, num_sec):
        raise NotImplementedError('wait_vm_stopped not implemented.')

    def create_vm(self, vm_name):
        raise NotImplementedError('create_vm not implemented.')

    def delete_vm(self, vm_name):
        raise NotImplementedError('delete_vm not implemented.')

    def restart_vm(self, vm_name):
        raise NotImplementedError('restart_vm not implemented.')

    def vm_status(self, vm_name):
        raise NotImplementedError('vm_status not implemented.')

    def is_vm_running(self, vm_name):
        raise NotImplementedError('is_vm_running not implemented.')

    def is_vm_stopped(self, vm_name):
        raise NotImplementedError('is_vm_stopped not implemented.')

    def is_vm_suspended(self, vm_name):
        raise NotImplementedError('is_vm_suspended not implemented.')

    def suspend_vm(self, vm_name):
        raise NotImplementedError('restart_vm not implemented.')

    def wait_vm_suspended(self, vm_name, num_sec):
        raise NotImplementedError('wait_vm_suspended not implemented.')

    def list_vm(self, **kwargs):
        raise NotImplementedError('list_vm not implemented.')

    def list_template(self):
        return [image.name for image in self.images]

    def list_flavor(self):
        raise NotImplementedError('list_flavor not implemented.')

    def list_network(self):
        return [network.name for network in self.networks]

    def list_host(self):
        return [node.name for node in self.nodes]

    def info(self):
        raise NotImplementedError('info not implemented.')

    def disconnect(self):
        pass

    def clone_vm(self, source_name, vm_name):
        raise NotImplementedError()

    def does_vm_exist(self, name):
        raise NotImplementedError()

    def deploy_template(self, template, *args, **kwargs):
        raise NotImplementedError()

    def current_ip_address(self, vm_name):
        raise NotImplementedError()

    def get_ip_address(self, vm_name):
        ""
        raise NotImplementedError()

    def remove_host_from_cluster(self, hostname):
        raise NotImplementedError()

    # TODO
